# -*- coding: utf-8 -*-


# Built-in
import copy


# Common
import numpy as np
import datastock as ds


# tofu
# from tofu import __version__ as __version__
from ._class00_Config import Config as Previous
from . import _class1_checks as _checks
from . import _class1_compute as _compute
from . import _class1_plot as _plot


__all__ = ['Plasma2D']


_WHICH_MESH = 'mesh'
_QUANT_R = 'R'
_QUANT_Z = 'Z'


# #############################################################################
# #############################################################################
#                           Plasma2D
# #############################################################################


class Plasma2D(Previous):

    _ddef = copy.deepcopy(ds.DataStock._ddef)
    _ddef['params']['ddata'].update({
        'bsplines': (str, ''),
    })
    _ddef['params']['dobj'] = None
    _ddef['params']['dref'] = None

    # _show_in_summary_core = ['shape', 'ref', 'group']
    _show_in_summary = 'all'
    _dshow = dict(Previous._dshow)

    _which_mesh = _WHICH_MESH
    _quant_R = _QUANT_R
    _quant_Z = _QUANT_Z

    def add_mesh(
        self,
        # rectangular mesh
        key=None,
        domain=None,
        res=None,
        R=None,
        Z=None,
        # triangular mesh
        knots=None,
        cents=None,
        # cropping
        crop_poly=None,
        thresh_in=None,
        remove_isolated=None,
        # direct addition of bsplines
        deg=None,
        **kwdargs,
    ):
        """ Add a mesh by key and domain / resolution

        Can create a rectangular or triangular mesh:
            - rectangular: provide (domain, res) or (R, Z)
                - domain:
                - res:
                - R:
                - Z:
            - triangular:
                - knots: (nknots, 2) array of (R, Z) coordinates
                - cents: (ncents, 3 or 4) array of int indices

        Can optionally be cropped by a closed polygon crop_poly, that can be:
            - a (2, N) np.narray of (R, Z) coordinates
            - a tuple (Config, key_struct) to designate a struct poly

        Args thresh_in and remove_isolated control the level of cropping:
            - thresh_in:
            - remove_isolated:

        If deg is provided, immediately adds a bsplines

        Example:
        --------
                >>> import tofu as tf
                >>> conf = tf.load_config('ITER')
                >>> mesh = tf.data.Plasma2D()
                >>> mesh.add_mesh(config=conf, res=0.1, deg=1)

        """

        # get domain, poly from crop_poly
        if crop_poly is not None:
            domain, poly = _checks._mesh2DRect_from_croppoly(
                crop_poly=crop_poly,
                domain=domain,
            )
        else:
            poly = None

        # check input data and get input dicts
        dref, ddata, dmesh = _checks._mesh2D_check(
            coll=self,
            # rectangular
            domain=domain,
            res=res,
            R=R,
            Z=Z,
            # triangular
            knots=knots,
            cents=cents,
            trifind=None,
            # key
            key=key,
        )

        # add kwdargs
        key = list(dmesh.keys())[0]
        dmesh[key].update(**kwdargs)

        # define dobj['mesh']
        dobj = {
            self._which_mesh: dmesh,
        }

        # update data source
        for k0, v0 in ddata.items():
            ddata[k0]['source'] = kwdargs.get('source')

        # update dicts
        self.update(dref=dref, ddata=ddata, dobj=dobj)

        # optional bspline
        if deg is not None:
            self.add_bsplines(key=key, deg=deg)

        # optional cropping
        c0 = (
            self.dobj[self._which_mesh][key]['type'] == 'rect'
            and poly is not None
        )
        if c0:
            self.crop(
                key=key,
                crop=poly,
                thresh_in=thresh_in,
                remove_isolated=remove_isolated,
            )

    def add_mesh_polar(
        self,
        # polar mesh
        key=None,
        radius=None,
        angle=None,
        # Defined on
        radius2d=None,
        angle2d=None,
        # optional special points coordinates vs time
        O_pts=None,         # computed if not provided
        X_pts=None,         # unused
        strike_pts=None,    # unused
        # res for contour discontinuity of angle2d
        res=None,
        # parameters
        radius_dim=None,
        radius_quant=None,
        radius_name=None,
        radius_units=None,
        angle_dim=None,
        angle_quant=None,
        angle_name=None,
        # direct addition of bsplines
        deg=None,
        **kwdargs,
    ):
        """ Add a 2d polar mesh

        For now only includes radial mesh
        radius has to be backed-up by:
            - a radius quantity from a pre-existing rect or tri mesh
            - a function

        """

        # check input data and get input dicts
        dref, ddata, dmesh = _checks._mesh2D_polar_check(
            coll=self,
            # polar
            radius=radius,
            angle=angle,
            radius2d=radius2d,
            angle2d=angle2d,
            # parameters
            radius_dim=radius_dim,
            radius_quant=radius_quant,
            radius_name=radius_name,
            radius_units=radius_units,
            angle_dim=angle_dim,
            angle_quant=angle_quant,
            angle_name=angle_name,
            # key
            key=key,
        )

        # add kwdargs
        key = list(dmesh.keys())[0]
        dmesh[key].update(**kwdargs)

        # update data source
        for k0, v0 in ddata.items():
            ddata[k0]['source'] = kwdargs.get('source')

        # special treatment of radius2d
        assert O_pts is None
        drefO, ddataO, kR, kZ = _compute.radius2d_special_points(
            coll=self,
            key=dmesh[key]['radius2d'],
            keym0=key,
            res=res,
        )
        dref.update(drefO)
        ddata.update(ddataO)
        dmesh[key]['pts_O'] = (kR, kZ)

        # define dobj['mesh']
        dobj = {
            self._which_mesh: dmesh,
        }

        # update dicts
        self.update(dref=dref, ddata=ddata, dobj=dobj)

        # special treatment of angle2d
        if dmesh[key]['angle2d'] is not None:
            drefa, ddataa, kR, kZ = _compute.angle2d_zone(
                coll=self,
                key=dmesh[key]['angle2d'],
                keyrad2d=dmesh[key]['radius2d'],
                key_ptsO=dmesh[key]['pts_O'],
                res=res,
                keym0=key,
            )

            # update dicts
            self.update(dref=drefa, ddata=ddataa)
            if 'azone' in self.get_lparam(self._which_mesh):
                self.set_param(
                    key=key,
                    param='azone',
                    value=(kR, kZ),
                    which=self._which_mesh,
                )
            else:
                self.add_param(
                    'azone', value={key: (kR, kZ)}, which=self._which_mesh,
                )

        # optional bspline
        if deg is not None:
            self.add_bsplines(key=key, deg=deg)

    # -----------------
    # bsplines
    # ------------------

    def add_bsplines(self, key=None, deg=None, angle=None):
        """ Add bspline basis functions on the chosen mesh """

        # --------------
        # check inputs

        keym, keybs, deg = _checks._mesh2D_bsplines(
            key=key,
            lkeys=list(self.dobj[self._which_mesh].keys()),
            deg=deg,
        )

        # --------------
        # get bsplines

        if self.dobj[self._which_mesh][keym]['type'] == 'rect':
            dref, ddata, dobj = _compute._mesh2DRect_bsplines(
                coll=self, keym=keym, keybs=keybs, deg=deg,
            )
        elif self.dobj[self._which_mesh][keym]['type'] == 'tri':
            dref, ddata, dobj = _compute._mesh2DTri_bsplines(
                coll=self, keym=keym, keybs=keybs, deg=deg,
            )
        else:
            dref, ddata, dobj = _compute._mesh2Dpolar_bsplines(
                coll=self, keym=keym, keybs=keybs, deg=deg, angle=angle,
            )

        # --------------
        # update dict and crop if relevant

        self.update(dobj=dobj, ddata=ddata, dref=dref)
        if self.dobj[self._which_mesh][keym]['type'] == 'rect':
            _compute.add_cropbs_from_crop(
                coll=self,
                keybs=keybs,
                keym=keym,
            )

    # -----------------
    # add_data
    # ------------------

    def update(
        self,
        dobj=None,
        ddata=None,
        dref=None,
        harmonize=None,
    ):
        """ Overload datastock update() method """

        # if ddata => check ref for bsplines
        if ddata is not None:
            for k0, v0 in ddata.items():
                (
                    ddata[k0]['ref'], ddata[k0]['data'],
                ) = _checks.add_data_meshbsplines_ref(
                    ref=v0['ref'],
                    data=v0['data'],
                    dmesh=self._dobj.get(self._which_mesh),
                    dbsplines=self._dobj.get('bsplines'),
                )

        # update
        super().update(
            dobj=dobj,
            ddata=ddata,
            dref=dref,
            harmonize=harmonize,
        )

        # assign bsplines
        if self._dobj.get('bsplines') is not None:
            for k0, v0 in self._ddata.items():
                lbs = [
                    k1 for k1, v1 in self._dobj['bsplines'].items()
                    if v1['ref'] == tuple([
                        rr for rr in v0['ref']
                        if rr in v1['ref']
                    ])
                ]
                if len(lbs) == 0:
                    pass
                elif len(lbs) == 1:
                    self._ddata[k0]['bsplines'] = lbs[0]
                else:
                    msg = f"Multiple nsplines:\n{lbs}"
                    raise Exception(msg)

        # assign diagnostic
        if self._dobj.get('camera') is not None:
            for k0, v0 in self._ddata.items():
                lcam = [
                    k1 for k1, v1 in self._dobj['camera'].items()
                    if v1['dgeom']['ref'] == tuple([
                        rr for rr in v0['ref']
                        if rr in v1['dgeom']['ref']
                    ])
                ]

                if len(lcam) == 0:
                    pass
                elif len(lcam) == 1:
                    self._ddata[k0]['camera'] = lcam[0]
                else:
                    msg = f"Multiple cameras:\n{lcam}"
                    raise Exception(msg)

    # -----------------
    # crop
    # ------------------

    def crop(self, key=None, crop=None, thresh_in=None, remove_isolated=None):
        """ Crop a mesh / bspline

        Uses:
            - a mask of bool for each mesh elements
            - a 2d (R, Z) closed polygon

        If applied on a bspline, cropping is double-checked to make sure
        all remaining bsplines have full support domain
        """
        crop, key, thresh_in = _compute.crop(
            coll=self,
            key=key,
            crop=crop,
            thresh_in=thresh_in,
            remove_isolated=remove_isolated,
        )

        # add crop data
        keycrop = f'{key}-crop'
        ref = tuple([
            self._ddata[k0]['ref'][0]
            for k0 in self._dobj[self._which_mesh][key]['cents']
        ])
        self.add_data(
            key=keycrop,
            data=crop,
            ref=ref,
            dim='bool',
            quant='bool',
        )

        # update obj
        self._dobj[self._which_mesh][key]['crop'] = keycrop
        self._dobj[self._which_mesh][key]['crop-thresh'] = thresh_in

        # also crop bsplines
        for k0 in self.dobj.get('bsplines', {}).keys():
            if self.dobj['bsplines'][k0][self._which_mesh] == key:
                _compute.add_cropbs_from_crop(coll=self, keybs=k0, keym=key)

    # -----------------
    # get data subset
    # ------------------

    def get_profiles2d(self):
        """ Return dict of profiles2d with associated bsplines as values """

        # dict of profiles2d
        dk = {
            k0: v0['bsplines']
            for k0, v0 in self._ddata.items()
            if v0['bsplines'] != ''
        }
        dk.update({k0: k0 for k0 in self._dobj['bsplines'].keys()})

        return dk

    # -------------------
    # get data time
    # -------------------

    def get_time(
        self,
        key=None,
        t=None,
        indt=None,
        ind_strict=None,
        dim=None,
    ):
        """ Return the time vector or time macthing indices

        hastime, keyt, reft, keyt, val, dind = self.get_time(key='prof0')

        Return
        ------
        hastime:    bool
            flag, True if key has a time dimension
        keyt:       None /  str
            if hastime and a time vector exists, the key to that time vector
        t:          None / np.ndarray
            if hastime
        dind:       dict, with:
            - indt:  None / np.ndarray
                if indt or t was provided, and keyt exists
                int indices of nearest matching times
            - indtu: None / np.ndarray
                if indt is returned, np.unique(indt)
            - indtr: None / np.ndarray
                if indt is returned, a bool (ntu, nt) array
            - indok: None / np.ndarray
                if indt is returned, a bool (nt,) array

        """

        if dim is None:
            dim = 'time'

        return self.get_ref_vector(
            key=key,
            values=t,
            indices=indt,
            ind_strict=ind_strict,
            dim=dim,
        )

    def get_time_common(
        self,
        keys=None,
        t=None,
        indt=None,
        ind_strict=None,
        dim=None,
    ):
        """ Return the time vector or time macthing indices

        hastime, hasvect, t, dind = self.get_time_common(
            keys=['prof0', 'prof1'],
            t=np.linspace(0, 5, 10),
        )

        Return
        ------
        hastime:        bool
            flag, True if key has a time dimension
        keyt:           None /  str
            if hastime and a time vector exists, the key to that time vector
        t:              None / np.ndarray
            if hastime
        indt:           None / np.ndarray
            if indt or t was provided, and keyt exists
            int indices of nearest matching times
        indtu:          None / np.ndarray
            if indt is returned, np.unique(indt)
        indt_reverse:   None / np.ndarray
            if indt is returned, a bool (ntu, nt) array

        """

        if dim is None:
            dim = 'time'

        return self.get_ref_vector_common(
            keys=keys,
            values=t,
            indices=indt,
            ind_strict=ind_strict,
            dim=dim,
        )

    # -----------------
    # indices
    # ------------------

    def select_ind(
        self,
        key=None,
        ind=None,
        elements=None,
        returnas=None,
        crop=None,
    ):
        """ Return ind for selected key (mesh or bspline) as:
                - tuple (default)
                - 'flat'

        Can covert one into the other
        """
        return _compute._select_ind(
            coll=self,
            key=key,
            ind=ind,
            elements=elements,
            returnas=returnas,
            crop=crop,
        )

    def select_mesh_elements(
        self,
        key=None,
        ind=None,
        elements=None,
        returnas=None,
        return_neighbours=None,
        crop=None,
    ):
        """ Return indices or values of selected knots / cent

        Can be used to convert tuple (R, Z) indices to flat (RZ,) indices
        Can return values instead of indices
        Can return indices / values of neighbourgs

        """
        # check key
        key = ds._generic_check._check_var(
            key, 'key',
            allowed=list(self.dobj.get('mesh', {}).keys()),
            types=str,
        )

        # get ind
        if self.dobj['mesh'][key]['type'] == 'rect':
            returnas_ind = tuple
        else:
            returnas_ind = bool

        ind = self.select_ind(
            key=key,
            ind=ind,
            elements=elements,
            returnas=returnas_ind,
            crop=crop,
        )

        return _compute._select_mesh(
            coll=self,
            key=key,
            ind=ind,
            elements=elements,
            returnas=returnas,
            return_neighbours=return_neighbours,
        )

    def select_bsplines(
        self,
        key=None,
        ind=None,
        returnas=None,
        return_cents=None,
        return_knots=None,
        crop=None,
    ):
        """ Return indices or values of selected knots / cent

        Can be used to convert tuple (R, Z) indices to flat (RZ,) indices
        Can return values instead of indices
        Can return indices / values of neighbourgs

        """
        return _compute._select_bsplines(
            coll=self,
            key=key,
            ind=ind,
            returnas=returnas,
            return_cents=return_cents,
            return_knots=return_knots,
            crop=crop,
        )

    # -----------------
    # Integration operators
    # ------------------

    def add_bsplines_operator(
        self,
        key=None,
        operator=None,
        geometry=None,
        crop=None,
        store=None,
        returnas=None,
        # specific to deg = 0
        centered=None,
        # to return gradR, gradZ, for D1N2 deg 0, for tomotok
        returnas_element=None,
    ):
        """ Get a matrix operator to compute an integral

        operator specifies the integrand:
            - 'D0N1': integral of the value
            - 'D0N2': integral of the squared value
            - 'D1N2': integral of the squared gradient
            - 'D2N2': integral of the squared laplacian

        geometry specifies in which geometry:
            - 'linear': linear geometry (cross-section = surface)
            - 'toroidal': toroildal geometry (cross-section = volumic slice)

        """

        (
            opmat, operator, geometry, dim, ref, crop,
            store, returnas, key,
        ) = _compute.get_bsplines_operator(
            self,
            key=key,
            operator=operator,
            geometry=geometry,
            crop=crop,
            store=store,
            returnas=returnas,
            # specific to deg = 0
            centered=centered,
            # to return gradR, gradZ, for D1N2 deg 0, for tomotok
            returnas_element=returnas_element,
        )

        # store
        if store is True:
            if operator == 'D1':
                name = f'{key}-{operator}-dR'
                if crop is True:
                    name = f'{name}-cropped'
                self.add_data(
                    key=name,
                    data=opmat[0],
                    ref=ref,
                    units='',
                    name=operator,
                    dim=dim,
                )
                name = f'{key}-{operator}-dZ'
                if crop is True:
                    name = f'{name}-cropped'
                self.add_data(
                    key=name,
                    data=opmat[1],
                    ref=ref,
                    units='',
                    name=operator,
                    dim=dim,
                )

            elif operator in ['D0N1', 'D0N2']:
                name = f'{key}-{operator}-{geometry}'
                if crop is True:
                    name = f'{name}-cropped'
                self.add_data(
                    key=name,
                    data=opmat,
                    ref=ref,
                    units='',
                    name=operator,
                    dim=dim,
                )
            elif operator == 'D1N2':
                name = f'{key}-{operator}-dR-{geometry}'
                if crop is True:
                    name = f'{name}-cropped'
                self.add_data(
                    key=name,
                    data=opmat[0],
                    ref=ref,
                    units='',
                    name=operator,
                    dim=dim,
                )
                name = f'{key}-{operator}-dZ-{geometry}'
                if crop is True:
                    name = f'{name}-cropped'
                self.add_data(
                    key=name,
                    data=opmat[1],
                    ref=ref,
                    units='',
                    name=operator,
                    dim=dim,
                )
            elif operator == 'D2N2':
                name = f'{key}-{operator}-d2R-{geometry}'
                if crop is True:
                    name = f'{name}-cropped'
                self.add_data(
                    key=name,
                    data=opmat[0],
                    ref=ref,
                    units='',
                    name=operator,
                    dim=dim,
                )
                name = f'{key}-{operator}-d2Z-{geometry}'
                if crop is True:
                    name = f'{name}-cropped'
                self.add_data(
                    key=name,
                    data=opmat[1],
                    ref=ref,
                    units='',
                    name=operator,
                    dim=dim,
                )
                name = f'{key}-{operator}-dRZ-{geometry}'
                if crop is True:
                    name = f'{name}-cropped'
                self.add_data(
                    key=name,
                    data=opmat[2],
                    ref=ref,
                    units='',
                    name=operator,
                    dim=dim,
                )
            else:
                msg = "Unknown opmat type!"
                raise Exception(msg)

        # return
        if returnas is True:
            return opmat, operator, geometry, dim, ref, crop

    # -----------------
    # interp tools
    # ------------------

    def get_sample_mesh(
        self,
        key=None,
        res=None,
        grid=None,
        mode=None,
        R=None,
        Z=None,
        DR=None,
        DZ=None,
        imshow=None,
    ):
        """ Return a sampled version of the chosen mesh """
        return _compute.sample_mesh(
            coll=self,
            key=key,
            res=res,
            grid=grid,
            mode=mode,
            R=R,
            Z=Z,
            DR=DR,
            DZ=DZ,
            imshow=imshow,
        )


    """
    def get_sample_bspline(self, key=None, res=None, grid=None, mode=None):
        return _compute.sample_bsplines(
            coll=self,
            key=key,
            res=res,
            grid=grid,
            mode=mode,
        )
    """

    def _check_qr12RPZ(
        self,
        quant=None,
        ref1d=None,
        ref2d=None,
        q2dR=None,
        q2dPhi=None,
        q2dZ=None,
        group1d=None,
        group2d=None,
    ):

        if group1d is None:
            group1d = self._group1d
        if group2d is None:
            group2d = self._group2d

        lc0 = [quant is None, ref1d is None, ref2d is None]
        lc1 = [q2dR is None, q2dPhi is None, q2dZ is None]
        if np.sum([all(lc0), all(lc1)]) != 1:
            msg = (
                "Please provide either (xor):\n"
                + "\t- a scalar field (isotropic emissivity):\n"
                + "\t\tquant : scalar quantity to interpolate\n"
                + "\t\t\tif quant is 1d, intermediate reference\n"
                + "\t\t\tfields are necessary for 2d interpolation\n"
                + "\t\tref1d : 1d reference field on which to interpolate\n"
                + "\t\tref2d : 2d reference field on which to interpolate\n"
                + "\t- a vector (R,Phi,Z) field (anisotropic emissivity):\n"
                + "\t\tq2dR :  R component of the vector field\n"
                + "\t\tq2dPhi: R component of the vector field\n"
                + "\t\tq2dZ :  Z component of the vector field\n"
                + "\t\t=> all components have the same time and mesh!\n"
            )
            raise Exception(msg)

        # Check requested quant is available in 2d or 1d
        if all(lc1):
            (
                idquant, idref1d, idref2d,
            ) = _compute._get_possible_ref12d(
                dd=self._ddata,
                key=quant, ref1d=ref1d, ref2d=ref2d,
                group1d=group1d,
                group2d=group2d,
            )
            idq2dR, idq2dPhi, idq2dZ = None, None, None
            ani = False
        else:
            idq2dR, msg = _compute._get_keyingroup_ddata(
                dd=self._ddata,
                key=q2dR, group=group2d, msgstr='quant', raise_=True,
            )
            idq2dPhi, msg = _compute._get_keyingroup_ddata(
                dd=self._ddata,
                key=q2dPhi, group=group2d, msgstr='quant', raise_=True,
            )
            idq2dZ, msg = _compute._get_keyingroup_ddata(
                dd=self._ddata,
                key=q2dZ, group=group2d, msgstr='quant', raise_=True,
            )
            idquant, idref1d, idref2d = None, None, None
            ani = True
        return idquant, idref1d, idref2d, idq2dR, idq2dPhi, idq2dZ, ani

    def _interp_pts2d_to_quant1d(
        self,
        pts=None,
        vect=None,
        t=None,
        quant=None,
        ref1d=None,
        ref2d=None,
        q2dR=None,
        q2dPhi=None,
        q2dZ=None,
        interp_t=None,
        interp_space=None,
        fill_value=None,
        Type=None,
        group0d=None,
        group1d=None,
        group2d=None,
        return_all=None,
    ):
        """ Return the value of the desired 1d quantity at 2d points

        For the desired inputs points (pts):
            - pts are in (X, Y, Z) coordinates
            - space interpolation is linear on the 1d profiles
        At the desired input times (t):
            - using a nearest-neighbourg approach for time

        """
        # Check inputs
        if group0d is None:
            group0d = self._group0d
        if group1d is None:
            group1d = self._group1d
        if group2d is None:
            group2d = self._group2d
        # msg = "Only 'nearest' available so far for interp_t!"
        # assert interp_t == 'nearest', msg

        # Check requested quant is available in 2d or 1d
        idquant, idref1d, idref2d, idq2dR, idq2dPhi, idq2dZ, ani = \
                self._check_qr12RPZ(
                    quant=quant, ref1d=ref1d, ref2d=ref2d,
                    q2dR=q2dR, q2dPhi=q2dPhi, q2dZ=q2dZ,
                    group1d=group1d, group2d=group2d,
                )

        # Check the pts is (3,...) array of floats
        idmesh = None
        if pts is None:
            # Identify mesh to get default points
            if ani:
                idmesh = [id_ for id_ in self._ddata[idq2dR]['ref']
                          if self._dref[id_]['group'] == group2d][0]
            else:
                if idref1d is None:
                    idmesh = [id_ for id_ in self._ddata[idquant]['ref']
                              if self._dref[id_]['group'] == group2d][0]
                else:
                    idmesh = [id_ for id_ in self._ddata[idref2d]['ref']
                              if self._dref[id_]['group'] == group2d][0]

            # Derive pts
            pts = self._get_pts_from_mesh(key=idmesh)

        pts = np.atleast_2d(pts)
        if pts.shape[0] != 3:
            msg = (
                "pts must be np.ndarray of (X,Y,Z) points coordinates\n"
                + "Can be multi-dimensional, but 1st dimension is (X,Y,Z)\n"
                + "    - Expected shape : (3,...)\n"
                + "    - Provided shape : {}".format(pts.shape)
            )
            raise Exception(msg)

        # Check t
        lc = [t is None, type(t) is str, type(t) is np.ndarray]
        assert any(lc)
        if lc[1]:
            assert t in self._ddata.keys()
            t = self._ddata[t]['data']

        # Interpolation (including time broadcasting)
        # this is the second slowest step (~0.08 s)
        func = self._get_finterp(
            idquant=idquant, idref1d=idref1d, idref2d=idref2d,
            idq2dR=idq2dR, idq2dPhi=idq2dPhi, idq2dZ=idq2dZ,
            idmesh=idmesh,
            interp_t=interp_t, interp_space=interp_space,
            fill_value=fill_value, ani=ani, Type=Type,
            group0d=group0d, group2d=group2d,
        )

        # Check vect of ani
        c0 = (
            ani is True
            and (
                vect is None
                or not (
                    isinstance(vect, np.ndarray)
                    and vect.shape == pts.shape
                )
            )
        )
        if c0:
            msg = (
                "Anisotropic field interpolation needs a field of local vect\n"
                + "  => Please provide vect as (3, npts) np.ndarray!"
            )
            raise Exception(msg)

        # This is the slowest step (~1.8 s)
        val, t = func(pts, vect=vect, t=t)

        # return
        if return_all is None:
            return_all = True
        if return_all is True:
            dout = {
                't': t,
                'pts': pts,
                'ref1d': idref1d,
                'ref2d': idref2d,
                'q2dR': idq2dR,
                'q2dPhi': idq2dPhi,
                'q2dZ': idq2dZ,
                'interp_t': interp_t,
                'interp_space': interp_space,
            }
            return val, dout
        else:
            return val

    def _interp_pts2d_to_quant1d(
        self,
        pts=None,
        vect=None,
        t=None,
        quant=None,
        ref1d=None,
        ref2d=None,
        q2dR=None,
        q2dPhi=None,
        q2dZ=None,
        interp_t=None,
        interp_space=None,
        fill_value=None,
        Type=None,
        group0d=None,
        group1d=None,
        group2d=None,
        return_all=None,
    ):
        """ Return the value of the desired 1d quantity at 2d points

        For the desired inputs points (pts):
            - pts are in (X, Y, Z) coordinates
            - space interpolation is linear on the 1d profiles
        At the desired input times (t):
            - using a nearest-neighbourg approach for time

        """
        # Check inputs
        if group0d is None:
            group0d = self._group0d
        if group1d is None:
            group1d = self._group1d
        if group2d is None:
            group2d = self._group2d
        # msg = "Only 'nearest' available so far for interp_t!"
        # assert interp_t == 'nearest', msg

        # Check requested quant is available in 2d or 1d
        idquant, idref1d, idref2d, idq2dR, idq2dPhi, idq2dZ, ani = \
                self._check_qr12RPZ(
                    quant=quant, ref1d=ref1d, ref2d=ref2d,
                    q2dR=q2dR, q2dPhi=q2dPhi, q2dZ=q2dZ,
                    group1d=group1d, group2d=group2d,
                )

        # Check the pts is (3,...) array of floats
        idmesh = None
        if pts is None:
            # Identify mesh to get default points
            if ani:
                idmesh = [id_ for id_ in self._ddata[idq2dR]['ref']
                          if self._dref[id_]['group'] == group2d][0]
            else:
                if idref1d is None:
                    idmesh = [id_ for id_ in self._ddata[idquant]['ref']
                              if self._dref[id_]['group'] == group2d][0]
                else:
                    idmesh = [id_ for id_ in self._ddata[idref2d]['ref']
                              if self._dref[id_]['group'] == group2d][0]

            # Derive pts
            pts = self._get_pts_from_mesh(key=idmesh)

        pts = np.atleast_2d(pts)
        if pts.shape[0] != 3:
            msg = (
                "pts must be np.ndarray of (X,Y,Z) points coordinates\n"
                + "Can be multi-dimensional, but 1st dimension is (X,Y,Z)\n"
                + "    - Expected shape : (3,...)\n"
                + "    - Provided shape : {}".format(pts.shape)
            )
            raise Exception(msg)

        # Check t
        lc = [t is None, type(t) is str, type(t) is np.ndarray]
        assert any(lc)
        if lc[1]:
            assert t in self._ddata.keys()
            t = self._ddata[t]['data']

        # Interpolation (including time broadcasting)
        # this is the second slowest step (~0.08 s)
        func = self._get_finterp(
            idquant=idquant, idref1d=idref1d, idref2d=idref2d,
            idq2dR=idq2dR, idq2dPhi=idq2dPhi, idq2dZ=idq2dZ,
            idmesh=idmesh,
            interp_t=interp_t, interp_space=interp_space,
            fill_value=fill_value, ani=ani, Type=Type,
            group0d=group0d, group2d=group2d,
        )

        # Check vect of ani
        c0 = (
            ani is True
            and (
                vect is None
                or not (
                    isinstance(vect, np.ndarray)
                    and vect.shape == pts.shape
                )
            )
        )
        if c0:
            msg = (
                "Anisotropic field interpolation needs a field of local vect\n"
                + "  => Please provide vect as (3, npts) np.ndarray!"
            )
            raise Exception(msg)

        # This is the slowest step (~1.8 s)
        val, t = func(pts, vect=vect, t=t)

        # return
        if return_all is None:
            return_all = True
        if return_all is True:
            dout = {
                't': t,
                'pts': pts,
                'ref1d': idref1d,
                'ref2d': idref2d,
                'q2dR': idq2dR,
                'q2dPhi': idq2dPhi,
                'q2dZ': idq2dZ,
                'interp_t': interp_t,
                'interp_space': interp_space,
            }
            return val, dout
        else:
            return val

    def interpolate_profile2d(
        # ressources
        self,
        # interpolation base, 1d or 2d
        key=None,
        # external coefs (optional)
        coefs=None,
        # interpolation points
        R=None,
        Z=None,
        radius=None,
        angle=None,
        grid=None,
        radius_vs_time=None,
        azone=None,
        # time: t or indt
        t=None,
        indt=None,
        indt_strict=None,
        # bsplines
        indbs=None,
        # parameters
        details=None,
        reshape=None,
        res=None,
        crop=None,
        nan0=None,
        val_out=None,
        imshow=None,
        return_params=None,
        # storing
        store=None,
        inplace=None,
    ):
        """ Interpolate desired profile2d (i.e.: data on bsplines)

        Interpolate:
            - key: a data on bsplines
            - coefs: external-provided set of coefs

        coefs can only be provided if:
            - details = False
            - key = keybs
            - coefs is a scalar or has shape = shapebs

        At points:
            - R:  R coordinates (np.ndarray or scalar)
            - Z:  Z coordinates (np.ndarray, same shape as R, or scalar)
            - grid: bool, if True mesh R x Z
            - indt: if provided, only interpolate at desired time indices

        With options:
            - details: bool, if True returns value for each bspline
            - indbs:   optional, select bsplines for which to interpolate
            - reshape: bool,
            - res:  optional, resolution to generate R and Z if they are None
            - crop: bool, whether to use the cropped mesh
            - nan0: value for out-of-mesh points
            - imshow: bool, whether to return as imshow (transpose)
            - return_params: bool, whether to return dict of input params

        """

        return _compute.interp2d(
            # ressources
            coll=self,
            # interpolation base, 1d or 2d
            key=key,
            # external coefs (optional)
            coefs=coefs,
            # interpolation points
            R=R,
            Z=Z,
            radius=radius,
            angle=angle,
            grid=grid,
            radius_vs_time=radius_vs_time,
            azone=azone,
            # time: t or indt
            t=t,
            indt=indt,
            indt_strict=indt_strict,
            # bsplines
            indbs=indbs,
            # parameters
            details=details,
            reshape=reshape,
            res=res,
            crop=crop,
            nan0=nan0,
            val_out=val_out,
            imshow=imshow,
            return_params=return_params,
            # storing
            store=store,
            inplace=inplace,
        )

    # TBF after polar meshes
    def interpolate_2dto1d(
        # resources
        self,
        # interpolation base
        key1d=None,
        key2d=None,
        # interpolation pts
        R=None,
        Z=None,
        grid=None,
        # parameters
        interp_t=None,
        fill_value=None,
        ani=False,
    ):

        return _compute.interp2dto1d(
            coll=self,
            key1d=key1d,
            key2d=key2d,
            R=R,
            Z=Z,
            grid=grid,
            crop=crop,
            nan0=nan0,
            return_params=return_params,
        )

    # -----------------
    # plotting
    # ------------------

    def plot_mesh(
        self,
        key=None,
        ind_knot=None,
        ind_cent=None,
        crop=None,
        color=None,
        dax=None,
        dmargin=None,
        fs=None,
        dleg=None,
        connect=None,
    ):

        return _plot.plot_mesh(
            coll=self,
            key=key,
            ind_knot=ind_knot,
            ind_cent=ind_cent,
            crop=crop,
            color=color,
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            dleg=dleg,
            connect=connect,
        )

    def plot_bsplines(
        self,
        key=None,
        indbs=None,
        indt=None,
        knots=None,
        cents=None,
        res=None,
        plot_mesh=None,
        val_out=None,
        nan0=None,
        cmap=None,
        dax=None,
        dmargin=None,
        fs=None,
        dleg=None,
    ):

        return _plot.plot_bspline(
            coll=self,
            key=key,
            indbs=indbs,
            indt=indt,
            knots=knots,
            cents=cents,
            res=res,
            plot_mesh=plot_mesh,
            val_out=val_out,
            nan0=nan0,
            cmap=cmap,
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            dleg=dleg,
        )

    def plot_profile2d(
        self,
        # inputs
        key=None,
        coefs=None,
        indt=None,
        res=None,
        # plot options
        vmin=None,
        vmax=None,
        cmap=None,
        dax=None,
        dmargin=None,
        fs=None,
        dcolorbar=None,
        dleg=None,
        # interactivity
        dinc=None,
        connect=None,
    ):
        return _plot.plot_profile2d(
            coll=self,
            # inputs
            key=key,
            coefs=coefs,
            indt=indt,
            res=res,
            # plot options
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            dcolorbar=dcolorbar,
            dleg=dleg,
            # interactivity
            dinc=dinc,
            connect=connect,
        )



"""
This module contains tests for tofu.geom in its structured version
"""

# Built-in
import os
import warnings

# Standard
import numpy as np
import matplotlib.pyplot as plt

# tofu-specific
from tofu import __version__
import tofu.utils as tfu
import tofu.geom as tfg
import tofu.data as tfd


_here = os.path.abspath(os.path.dirname(__file__))
VerbHead = 'tofu.data.test_03_core'


#######################################################
#
#     Setup and Teardown
#
#######################################################

def setup_module():
    print("") # this is to get a newline after the dots
    LF = os.listdir(_here)
    LF = [lf for lf in LF if all([ss in lf for ss in ['TFD_','Test','.npz']])]
    LF = [lf for lf in LF if not lf[lf.index('_Vv')+2:lf.index('_U')]==__version__]
    print("Removing the following previous test files:")
    print (LF)
    for lf in LF:
        os.remove(os.path.join(_here,lf))
    #print("setup_module before anything in this file")

def teardown_module():
    #os.remove(VesTor.Id.SavePath + VesTor.Id.SaveName + '.npz')
    #os.remove(VesLin.Id.SavePath + VesLin.Id.SaveName + '.npz')
    #print("teardown_module after everything in this file")
    #print("") # this is to get a newline
    LF = os.listdir(_here)
    LF = [lf for lf in LF if all([ss in lf for ss in ['TFD_','Test','.npz']])]
    LF = [lf for lf in LF if lf[lf.index('_Vv')+2:lf.index('_U')]==__version__]
    print("Removing the following test files:")
    print (LF)
    for lf in LF:
        os.remove(os.path.join(_here,lf))
    pass


#######################################################
#
#     Creating Ves objects and testing methods
#
#######################################################

def emiss(pts, t=None):
    R = np.hypot(pts[0,:],pts[1,:])
    Z = pts[2,:]
    r = np.hypot(R-2.4, Z)
    e0 = np.exp(-r**2/0.25)
    e1 = np.exp(-r**2/0.1)
    e2 = np.exp(-(r-0.4)**2/0.1)
    if t is None:
        emiss = e0 + 0.1*e1
    else:
        emiss = (e0[None,:]
                 + 0.5*np.cos(t)[:,None]*e1[None,:]
                 + 0.1*np.sin(2*t)[:,None]*e2[None,:])
    return emiss




class Test01_DataCam12D(object):


    @staticmethod
    def _create_cams(nch, lconf, ld, SavePath='./'):
        c0 = tfg.utils.create_CamLOS1D(P=[4.5,0,0], F=0.1, N12=nch, D12=0.05,
                                       angs=[-np.pi,np.pi/10.,0.],
                                       config=lconf[0],
                                       Diag='Test', Name='Test',
                                       SavePath=SavePath)
        c1 = tfg.utils.create_CamLOS2D(P=[4.5,0,0], F=0.1,
                                       N12=[int(1.5*nch),nch],
                                       D12=[0.075,0.05],
                                       angs=[-np.pi,np.pi/10.,0.],
                                       config=lconf[1],
                                       Diag='Test', Name='Test',
                                       SavePath=SavePath)
        return [c0, c1]

    @classmethod
    def setup_class(cls, nch=30, nt=50, SavePath='./', verb=False):

        # time vector
        t = np.linspace(0, 10, nt)

        # Configs
        conf0 = tfg.utils.create_config(case='B2')
        conf1 = tfg.utils.create_config(case='B3')

        # dchans and cams
        d0 = dict(Name=['C0-{0}'.format(ii) for ii in range(0,nch)])
        d1 = dict(Name=['C1-{0}'.format(ii) for ii in range(0,nch)])
        lc = cls._create_cams(nch, [conf0, conf1], [d0, d1], SavePath=SavePath)

        # -------
        # dextra
        nteq = nt // 2
        teq = np.linspace(t.min(), t.max(), nteq)
        teq2 = np.copy(teq) - 0.01
        Ax = np.array([2.4+0.1*np.cos(teq2), 0.1*np.sin(teq2)]).T
        Ax2 = np.array([2.4+0.1*np.cos(teq2/2.), 0.1*np.sin(teq2/2.)]).T
        Sep = (Ax[:,:,None]
               + 0.4*np.array([[-1,1,1,-1],[-1,-1,1,1]])[None,:,:])
        Sep2 = (Ax2[:,:,None]
                + 0.3*np.array([[-1,1,1,-1],[-1,-1,1,1]])[None,:,:])

        n1, n2 = 40, 60
        x1, x2 = np.linspace(2,3,n1), np.linspace(-0.8,0.8,n2)
        dx1, dx2 = (x1[1]-x1[0])/2., (x2[1]-x2[0])/2
        extent = (x1[0]-dx1, x1[-1]+dx1, x2[0]-dx2, x2[-1]+dx2)
        pts = np.array([np.tile(x1,n2), np.zeros((n1*n2,)), np.repeat(x2,n1)])
        emis = emiss(pts, t=teq2).reshape(nteq, n2, n1)
        dextra0 = {'pouet':{'t':teq, 'c':'k', 'data':np.sin(teq),
                            'units':'a.u.' , 'label':'pouet'},
                   'Ax':{'t':teq2, 'data2D':Ax},
                   'Sep':{'t':teq2, 'data2D':Sep},
                   'map':{'t':teq2, 'data2D':emis, 'extent':extent}}
        dextra1 = {'pouet':{'t':teq, 'c':'k', 'data':np.cos(teq),
                            'units':'a.u.' , 'label':'pouet'},
                   'Ax':{'t':teq2, 'data2D':Ax2},
                   'Sep':{'t':teq2, 'data2D':Sep2}}

        # -------
        # signal as Data from lcams
        lm = ['sum', 'simps']
        lData = [lc[ii].calc_signal(emiss, t=t,
                                    res=0.01, method=lm[ii], plot=False)[0]
                 for ii in range(0,len(lc))]

        # Adding concatenated sig / data and without lcam
        sig = np.concatenate([dd.data for dd in lData[:2]], axis=1)
        lData += [tfd.DataCam1D(data=sig, Name='All',
                                Diag='Test', Exp=conf0.Id.Exp, config=conf0)]
        dX12 = lc[1].dX12
        lData += [tfd.DataCam2D(data=lData[1].data, dX12=dX12, Name='c1nocam',
                                Diag='Test', Exp=conf0.Id.Exp)]

        # Setting dchans
        for ii in range(0,len(lData)):
            if ii % 2 == 0:
                lData[ii].set_dchans({'Name':['c%s'%jj for jj in
                                              range(0,lData[ii].nch)]})

        # Setting dextra
        for ii in range(0,len(lData)):
            de = dextra0 if ii % 2 == 0 else dextra1
            lData[ii].set_dextra(dextra=de)

        # Storing
        cls.lobj = lData
        cls.t = t

        # Saving for intermediate use
        lpfe = []
        for oo in cls.lobj:
            if oo._dgeom['config'] is not None:
                lpfe.append( oo._dgeom['config'].save(return_pfe=True,
                                                      verb=verb) )
            if oo._dgeom['lCam'] is not None:
                for cc in oo._dgeom['lCam']:
                    lpfe.append( cc.save(return_pfe=True, verb=verb) )

        cls.lpfe = lpfe

    @classmethod
    def setup_method(self):
        pass
    def teardown_method(self):
        pass

    @classmethod
    def teardown_class(cls):
        for pfe in set(cls.lpfe):
            os.remove(pfe)
        pass

    def test01_set_dchans(self):
        for ii in range(0,len(self.lobj)):
            oo = self.lobj[ii]
            out0 = oo.dchans()
            if out0 is not None:
                assert type(out0) is dict
                lk = list(out0.keys())
                if len(lk)>0:
                    out1 = oo.dchans(key=lk[0])
                    assert type(out1) is np.ndarray
                    dch2 = {'dch2':['abc' for ii in range(0,oo.ddataRef['nch'])]}
                    oo.set_dchans(dch2, method='update')
                assert all([len(out0[ss])==oo.ddataRef['nch'] for ss in lk])

    def test02_set_dextra(self):
        for ii in range(0,len(self.lobj)):
            oo = self.lobj[ii]
            out0 = oo.dextra
            if out0 is not None:
                assert type(out0) is dict
                t = np.linspace(0,10,10)
                dd = {'pouet': {'t':t, 'data':t,
                                'label':'pouet', 'units':'pouet units'}}
                oo.set_dextra(dd, method='update')

    def test03_select_t(self):
        for oo in self.lobj:
            ind = oo.select_t(t=None, out=bool)
            assert ind.sum()==oo.ddataRef['nt']
            ind = oo.select_t(t=5, out=bool)
            assert ind.sum() == 1
            ind = oo.select_t(t=[1,4], out=bool)
            assert np.all((oo.t[ind]>=1.) & (oo.t[ind]<=4))

    def test04_select_ch(self):
        for oo in self.lobj:
            if oo.dgeom['lCam'] is not None:
                name = [(ii, k) for ii, k in
                        enumerate(oo.config.dStruct['lorder'])
                        if 'Ves' in k or 'PlasmaDomain' in k]
                # assert len(name) == 1  # There can be several Ves now
                ind = oo.select_ch(touch=name[0][1], out=bool)
                assert ind.sum() > 0, (ind.sum(), ind)
                assert np.allclose(ind, oo.select_ch(touch=name[0][0],
                                                     out=bool))
            if len(oo.dchans().keys()) > 0:
                ind = oo.select_ch(key='Name', val=['c0','c10'],
                                   log='any', out=bool)
                assert ind.sum() == 2, (ind.sum(), ind)

    def test05_set_dtreat_indt(self):
        for oo in self.lobj:
            oo.set_dtreat_indt(t=[2,3])
            assert np.all((oo.t>=2) & (oo.t<=3))
            oo.set_dtreat_indt(indt=list(range(0,min(4,oo.ddataRef['nt']))))
            assert oo.nt == 4 or oo.nt==1
            oo.set_dtreat_indt()
            assert oo.nt == oo.ddataRef['nt']

    def test06_set_dtreat_indch(self):
        for oo in self.lobj:
            oo.set_dtreat_indch(indch = range(0,10))
            assert oo.dtreat['indch'].sum() == 10

    def test07_set_dtreat_mask(self):
        for oo in self.lobj:
            # Re-initialise
            oo.set_dtreat_indch()
            # set mask
            mask = np.arange(0,oo.ddataRef['nch'],10)
            oo.set_dtreat_mask(ind=mask, val=np.nan)
            nbnan = np.sum(np.any(np.isnan(oo.data), axis=0))
            assert nbnan >= mask.size, [oo.ddataRef['nch'], nbnan]

    def test08_dtreat_set_data0(self):
        for oo in self.lobj:
            # Re-initialise
            oo.set_dtreat_indt()
            oo.set_dtreat_mask()

            oo.set_dtreat_data0( data0 = oo.data[0,:] )
            assert oo.dtreat['data0-indt'] is None
            assert oo.dtreat['data0-Dt'] is None
            assert np.allclose(oo.data[0,:],0.), oo.data[0,:]

            oo.set_dtreat_data0(indt=[1,2,6,8,9])
            assert oo.dtreat['data0-indt'].sum() == 5
            assert oo.dtreat['data0-data'].size == oo.ddataRef['nch']

            oo.set_dtreat_data0(Dt=[2,3])
            assert oo.dtreat['data0-Dt'][0] >= 2. and oo.dtreat['data0-Dt'][1] <= 3.
            assert oo.dtreat['data0-data'].size == oo.ddataRef['nch']

            oo.set_dtreat_data0()
            assert oo.dtreat['data0-data'] is None
            assert np.allclose(oo.data, oo.ddataRef['data'])

    def test09_dtreat_set_interp_indt(self):
        for oo in self.lobj:
            ind = np.arange(0,oo.nt,10)
            oo.set_dtreat_interp_indt( ind )
            assert oo._dtreat['interp-indt'].sum() == ind.size

            ind = dict([(ii, np.arange(0,oo.nt,5)) for ii in range(0,oo.nch,3)])
            oo.set_dtreat_interp_indt( ind )
            assert type(oo._dtreat['interp-indt']) is dict

            oo.set_dtreat_interp_indt()
            assert oo._dtreat['interp-indt'] is None

    def test10_dtreat_set_interp_indch(self):
        for oo in self.lobj:
            ind = np.arange(0, oo.nch, 10, dtype=int)
            oo.set_dtreat_interp_indch( ind )
            assert oo._dtreat['interp-indch'].sum() == ind.size

            ind = dict([(ii, np.arange(0,oo.nch,5)) for ii in range(0,oo.nt,3)])
            oo.set_dtreat_interp_indch( ind )
            assert type(oo._dtreat['interp-indch']) is dict

            oo.set_dtreat_interp_indch()
            assert oo._dtreat['interp-indch'] is None

    def test11_streat_set_dfit(self):
        for oo in self.lobj:
            oo.set_dtreat_dfit()

    def test12_streat_set_interpt(self):
        t = np.linspace(self.t[0]-0.1, self.t[-1]+0.5, 100)
        for oo in self.lobj:
            oo.set_dtreat_interpt(t)

    def test13_clear_ddata(self):
        for ii in range(0,len(self.lobj)):
            if ii%2 == 0:
                self.lobj[ii].clear_ddata()

    def test14_clear_dtreat(self):
        for ii in range(0,len(self.lobj)):
            if ii%2 == 1:
                self.lobj[ii].clear_dtreat(force=True)


    def test15_plot(self):
        for oo in self.lobj:
            kh = oo.plot(key=None, ntMax=4, nchMax=2, fs=None,
                         dmargin=dict(left=0.06, right=0.9),
                         wintit='test', tit='AHAH')
        plt.close('all')

    def test16_compare(self):
        for oo in self.lobj:
            kh = oo.plot_compare(oo)
        plt.close('all')

    def test17_plot_combine(self):
        for ii in range(1,len(self.lobj)):
            kh = self.lobj[ii].plot_combine(self.lobj[ii-1])
        plt.close('all')

    def test18_spectrogram(self):
        for oo in self.lobj:
            kh = oo.plot_spectrogram(warn=False)
        plt.close('all')

    def test19_plot_svd(self):
        for oo in self.lobj:
            kh = oo.plot_svd()
        plt.close('all')

    def test20_copy_equal(self):
        for oo in self.lobj:
            obj = oo.copy()
            assert obj == oo

    def test21_get_nbytes(self):
        for oo in self.lobj:
            nb, dnb = oo.get_nbytes()

    def test22_strip_nbytes(self, verb=False):
        lok = self.lobj[0].__class__._dstrip['allowed']
        nb = np.full((len(lok),), np.nan)
        for oo in self.lobj:
            for ii in lok:
                oo.strip(ii, verb=verb)
                nb[ii] = oo.get_nbytes()[0]
            assert np.all(np.diff(nb)<=0.), nb
            for ii in lok[::-1]:
                oo.strip(ii, verb=verb)

    def test23_saveload(self, verb=False):
        for oo in self.lobj:
            pfe = oo.save(deep=False, verb=verb, return_pfe=True)
            obj = tfu.load(pfe, verb=verb)
            # Just to check the loaded version works fine
            assert oo == obj
            os.remove(pfe)





class Test02_DataCam12DSpectral(Test01_DataCam12D):

    @classmethod
    def setup_class(cls, nch=30, nt=50, SavePath='./', verb=False):

        # time vector
        t = np.linspace(0, 10, nt)

        # Configs
        conf0 = tfg.utils.create_config(case='B2')
        conf1 = tfg.utils.create_config(case='B3')

        # dchans and cams
        d0 = dict(Name=['C0-{0}'.format(ii) for ii in range(0,nch)])
        d1 = dict(Name=['C1-{0}'.format(ii) for ii in range(0,nch)])
        lc = cls._create_cams(nch, [conf0, conf1], [d0, d1], SavePath=SavePath)

        # -------
        # dextra
        nteq = nt // 2
        teq = np.linspace(t.min(), t.max(), nteq)
        teq2 = np.copy(teq) - 0.01
        Ax = np.array([2.4+0.1*np.cos(teq2), 0.1*np.sin(teq2)]).T
        Ax2 = np.array([2.4+0.1*np.cos(teq2/2.), 0.1*np.sin(teq2/2.)]).T
        Sep = (Ax[:,:,None]
               + 0.4*np.array([[-1,1,1,-1],[-1,-1,1,1]])[None,:,:])
        Sep2 = (Ax2[:,:,None]
                + 0.3*np.array([[-1,1,1,-1],[-1,-1,1,1]])[None,:,:])

        n1, n2 = 40, 60
        x1, x2 = np.linspace(2,3,n1), np.linspace(-0.8,0.8,n2)
        dx1, dx2 = (x1[1]-x1[0])/2., (x2[1]-x2[0])/2
        extent = (x1[0]-dx1, x1[-1]+dx1, x2[0]-dx2, x2[-1]+dx2)
        pts = np.array([np.tile(x1,n2), np.zeros((n1*n2,)), np.repeat(x2,n1)])
        emis = emiss(pts, t=teq2).reshape(nteq, n2, n1)
        dextra0 = {'pouet':{'t':teq, 'c':'k', 'data':np.sin(teq),
                            'units':'a.u.' , 'label':'pouet'},
                   'Ax':{'t':teq2, 'data2D':Ax},
                   'Sep':{'t':teq2, 'data2D':Sep},
                   'map':{'t':teq2, 'data2D':emis, 'extent':extent}}
        dextra1 = {'pouet':{'t':teq, 'c':'k', 'data':np.cos(teq),
                            'units':'a.u.' , 'label':'pouet'},
                   'Ax':{'t':teq2, 'data2D':Ax2},
                   'Sep':{'t':teq2, 'data2D':Sep2}}

        # lamb
        nlamb = 100
        lamb = np.linspace(10,20,nlamb)
        flamb = np.exp(-(lamb-12)**2/0.1) + 0.4*np.exp(-(lamb-16)**2/0.5)

        # -------
        # signal as Data from lcams
        lm = ['sum', 'simps']
        lData = [None for ii in range(0,len(lc))]
        for ii in range(0,len(lc)):
            sig = lc[ii].calc_signal(emiss, t=t, res=0.01, method=lm[ii],
                                     plot=False, returnas=np.ndarray)[0]
            sig = sig[:,:,None]*flamb[None,None,:]
            cla = eval('tfd.DataCam%sDSpectral'%('2' if lc[ii]._is2D() else '1'))
            data = cla(data=sig, Name='All', Diag='Test',
                       Exp=conf0.Id.Exp, lCam=lc[ii], t=t,
                       lamb=lamb)
            lData[ii] = data

        # Setting dchans
        for ii in range(0,len(lData)):
            if ii % 2 == 0:
                lData[ii].set_dchans({'Name':['c%s'%jj for jj in
                                              range(0,lData[ii].nch)]})

        # Setting dextra
        for ii in range(0,len(lData)):
            de = dextra0 if ii % 2 == 0 else dextra1
            lData[ii].set_dextra(dextra=de)

        # Storing
        cls.lobj = lData
        cls.t = t

        # Saving for intermediate use
        lpfe = []
        for oo in cls.lobj:
            if oo._dgeom['config'] is not None:
                lpfe.append( oo._dgeom['config'].save(return_pfe=True,
                                                      verb=verb) )
            if oo._dgeom['lCam'] is not None:
                for cc in oo._dgeom['lCam']:
                    lpfe.append( cc.save(return_pfe=True, verb=verb) )

        cls.lpfe = lpfe

    def test08_dtreat_set_data0(self):
        for oo in self.lobj:
            # Re-initialise
            oo.set_dtreat_indt()
            oo.set_dtreat_mask()

            oo.set_dtreat_data0( data0 = oo.data[0,:,:] )
            assert oo.dtreat['data0-indt'] is None
            assert oo.dtreat['data0-Dt'] is None
            assert np.allclose(oo.data[0,:,:],0.), oo.data[0,:,:]

            oo.set_dtreat_data0(indt=[1,2,6,8,9])
            assert oo.dtreat['data0-indt'].sum() == 5
            assert oo.dtreat['data0-data'].shape == (oo.ddataRef['nch'],
                                                     oo.ddataRef['nlamb'])

            oo.set_dtreat_data0(Dt=[2,3])
            assert oo.dtreat['data0-Dt'][0] >= 2. and oo.dtreat['data0-Dt'][1] <= 3.
            assert oo.dtreat['data0-data'].shape == (oo.ddataRef['nch'],
                                                     oo.ddataRef['nlamb'])

            oo.set_dtreat_data0()
            assert oo.dtreat['data0-data'] is None
            assert np.allclose(oo.data, oo.ddataRef['data'])

    def test17_plot_combine(self):
        pass

    def test18_spectrogram(self):
        pass

    def test19_plot_svd(self):
        pass





"""
    def test12_operators(self):
        o0 = self.lobj[-1]
        o1 = 100.*(o0-0.1*o0)

    def test15_combine(self):
        if self.__class__ is Test02_Data2D:
            return
        connect = (hasattr(plt.get_current_fig_manager(),'toolbar')
                   and plt.get_current_fig_manager().toolbar is not None)
        o0 = self.lobj[0]
        for ii in range(1,len(self.lobj)):
            oo = self.lobj[ii]
            KH = oo.plot_combine(o0, connect=connect)
        plt.close('all')

"""



"""
class Test02_Data2D(Test01_Data1D):
    @classmethod
    def setup_class(cls):
        thet = np.linspace(0,2.*np.pi,100)
        P = np.array([2.4 + 0.8*np.cos(thet),0.8*np.sin(thet)])
        V = tfg.Ves(Name='Test', Poly=P, Exp='Dummy', SavePath=here)
        N = 5
        Ds, us = tfg.utils.compute_CamLOS2D_pinhole([3.5,0.,0.], 0.1,
                                                    (0.05,0.05), (N,N),
                                                    angs=None,
                                                    nIn=[-1,0.,0.],
                                                    VType='Tor',
                                                    return_Du=True)
        d0 = dict(Name=['C0-{0}'.format(ii) for ii in range(0,N**2)])
        d1 = dict(Name=['C1-{0}'.format(ii) for ii in range(0,N**2)])
        config = tfg.Config(Name="Conf", lStruct=[V])
        C0 = tfg.CamLOS2D(Name="Dummy", dgeom=(Ds,us), config=config,
                          Exp=config.Id.Exp, Diag='Test', dchans=d0, SavePath=here)
        V.save()
        C0.save()
        t = np.linspace(0,10,20)
        sig00 = C0.calc_signal(emiss, t=None, res=0.01, method='sum',
                               plot=False)
        sig01 = C0.calc_signal(emiss, t=t, res=0.01, method='sum',
                               plot=False)
        cls.lobj = [tfd.Data2D(sig00, Id='0', SavePath=here),
                    tfd.Data2D(sig01, t=t, Id='1', SavePath=here),
                    tfd.Data2D(sig01, t=t, Ves=V, LStruct=C0.LStruct,
                               Id='1', SavePath=here),
                    tfd.Data2D(sig00, LCam=C0, Id='2', SavePath=here),
                    tfd.Data2D(sig01, t=t, LCam=C0, Id='3', SavePath=here)]

    def test08_plot(self):
        for ii in range(0,len(self.lobj)):
            oo = self.lobj[ii]
            if oo._X12 is not None and oo.geom is not None:
                oo.set_indch()
                KH = oo.plot(key=None, Max=None, fs=None,
                             invert=True, vmin=0, wintit='test', tit='AHAH',
                             dmargin=dict(left=0.05,right=0.9))
                KH = oo.plot(key='Name', Max=2, fs=(13,5),
                             normt=True, dmargin=None)
        plt.close('all')
"""



#!/usr/bin/env python

# Built-in
import sys
import os
import warnings


_HERE = os.path.abspath(os.path.dirname(__file__))


# import parser dict
sys.path.insert(1, _HERE)
from _dparser import _DPARSER
_ = sys.path.pop(1)


###################################################
###################################################
#       function
###################################################


def get_version(verb=None, envvar=None,
                path=None, warn=None, force=None, ddef=None):
    """ Print tofu version and / or store in environment variable """

    # --------------
    # Check inputs
    kwd = locals()
    for k0 in set(ddef.keys()).intersection(kwd.keys()):
        if kwd[k0] is None:
            kwd[k0] = ddef[k0]
    verb, envvar, path = kwd['verb'], kwd['envvar'], kwd['path']
    warn, force = kwd['warn'], kwd['force']

    # verb, warn, force
    dbool = {'verb': verb, 'warn': warn, 'force': force}
    for k0, v0 in dbool.items():
        if v0 is None:
            dbool[k0] = ddef[k0]
        if not isinstance(dbool[k0], bool):
            msg = ("Arg {} must be a bool\n".format(k0)
                   + "\t- provided: {}".format(dbool[k0]))
            raise Exception(msg)

    # envvar
    if isinstance(envvar, bool):
        if envvar is True:
            envvar = ddef['name']
    elif isinstance(envvar, str):
        envvar = envvar.upper()
    else:
        msg = ("Arg envvar must be either:\n"
               + "\t- None:  set to default ({})\n".format(ddef['envvar'])
               + "\t- False: no setting of environment variable\n"
               + "\t- True:  set default env. var. ({})\n".format(ddef['name'])
               + "\t- str:   set provided env. variable\n\n"
               + " you provided: {}".format(envvar))
        raise Exception(msg)

    if envvar is not False and envvar in os.environ.keys():
        if dbool['warn'] is True:
            msg = ("Chosen environment variable name already exists:\n"
                   + "\t- {}: {}\n".format(envvar, os.environ[envvar])
                   + "  => Are you sure you want to change it?\n"
                   + "       use force=True to overwrite {}\n".format(envvar)
                   + "       use warn=False to disable this warning")
            warnings.warn(msg)

    # --------------
    # Fetch version from git tags, and write to version.py
    # Also, when git is not available (PyPi package), use stored version.py
    pfe = os.path.join(path, 'version.py')
    if not os.path.isfile(pfe):
        msg = ("It seems your current tofu install has no version.py:\n"
               "\t- looked for: {}".format(pfe))
        raise Exception(msg)

    # --------------
    # Read file
    with open(pfe, 'r') as fh:
        version = fh.read().strip().split("=")[-1].replace("'", '')
    version = version.lower().replace('v', '').replace(' ', '')

    # --------------
    # Outputs
    if dbool['verb'] is True:
        print(version)
    if envvar is not False:
        c0 = ((envvar in os.environ.keys() and dbool['force'] is True)
              or envvar not in os.environ.keys())
        if c0 is True:
            os.environ[envvar] = version


###################################################
###################################################
#       bash call (main)
###################################################


def main():
    # Parse input arguments

    # Parse arguments
    ddef, parser = _DPARSER['version']()

    # Parse arguments
    args = parser.parse_args()

    # Call function
    get_version(ddef=ddef, **dict(args._get_kwargs()))


if __name__ == '__main__':
    main()





from . import tests00_root
from . import tests01_geom
from . import tests02_data
from . import tests03_openadas
from . import tests04_spectro
from . import tests05_nist
from . import tests06_mesh
from . import tests07_inversions
from . import tests08_diagnostics
from . import tests09_tutorials



# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np
import scipy.interpolate as scpinterp
import scipy.spatial as scpspace
from matplotlib.tri import Triangulation as mplTri
import datastock as ds


# specific
from . import _class1_checks as _checks
from . import _class1_bsplines_operators_tri
from . import _class1_bsplines_rect as _mbr


# #############################################################################
# #############################################################################
#                       BivariateSplineRect - scipy subclass
# #############################################################################


class BivariateSplineTri(scpinterp.BivariateSpline):
    """ Subclass for tofu

    Defined from knots (unique) and deg
    coefs set to 1 by default

    """

    def __init__(
        self,
        knotsR=None,
        knotsZ=None,
        cents=None,
        trifind=None,
        deg=None,
    ):
        """ Class handling triangular bsplines """

        # ------------
        # check inputs

        knots = np.array([knotsR, knotsZ]).T
        cents, knots = _checks._mesh2DTri_conformity(
            knots=knots, cents=cents, key='class',
        )
        cents = _checks._mesh2DTri_clockwise(
            knots=knots, cents=cents, key='class',
        )

        # check trifinder
        if trifind is None:
            mpltri = mplTri(knotsR, knotsZ, cents)
            trifind = mpltri.get_trifinder()

        # deg
        deg = ds._generic_check._check_var(
            deg, 'deg',
            types=int,
            default=2,
            allowed=[0, 1, 2, 3],
        )

        if deg not in [0, 1]:
            msg = "Only deg=0 implemented for triangular meshes so far!"
            raise NotImplementedError(msg)

        self.knotsR = knotsR
        self.knotsZ = knotsZ
        self.nknots = knots.shape[0]
        self.cents = cents
        self.ncents = cents.shape[0]
        self.deg = deg
        self.trifind = trifind

        # ------------
        # get neigh cents per knot

        self.cents_per_knots = self._get_cents_per_knots()

        # ------------
        # get height per cent /  knot

        self.heights = self._get_heights_per_centsknots()

        # ------------
        # nbsplines

        if deg == 0:
            nbs = self.ncents
        elif deg == 1:
            nbs = self.nknots
        elif deg == 2:
            raise NotImplementedError()

        self.nbs = nbs
        self.shapebs = (nbs,)
        self.coefs = np.ones((nbs,), dtype=float)

    def _get_cents_per_knots(self):
        """ Return a (nknots, nmax) array of int indices

        Array contains -1 where there is no cent anymore
        """

        out = [
            np.any(self.cents == ii, axis=1).nonzero()[0]
            for ii in range(self.nknots)
        ]
        nmax = np.array([oo.size for oo in out])

        cents_per_knots = -np.ones((self.nknots, nmax.max()), dtype=int)
        for ii in range(self.nknots):
            cents_per_knots[ii, :nmax[ii]] = out[ii]
        return cents_per_knots

    def _get_heights_per_centsknots(self):
        """ Return the height of each knot in each cent

        Returnad as (ncents, 3) array, like cents
        """

        R = self.knotsR[self.cents]
        Z = self.knotsZ[self.cents]

        heights = np.full(self.cents.shape, np.nan)

        for iref, (i0, i1) in enumerate([(1, 2), (2, 0), (0, 1)]):
            base = np.sqrt(
                (R[:, i1] - R[:, i0])**2 + (Z[:, i1] - Z[:, i0])**2
            )
            heights[:, iref] = np.abs(
                (R[:, i0] - R[:, iref])*(Z[:, i1] - Z[:, iref])
                - (Z[:, i0] - Z[:, iref])*(R[:, i1] - R[:, iref])
            ) / base

        return heights

    def get_heights_per_centsknots_pts(self, x, y):
        """ Return the height of each knot in each cent

        Returnad as (ncents, 3) array, like cents

        OPTIMIZATION POSSIBLE FOR EV_DETAILS BY TAKING INDBS ASINPUT ARG !!!
        """

        if x.shape != y.shape:
            msg = "Arg x and y must hae the same shape!"
            raise Exception(msg)

        R = self.knotsR[self.cents]
        Z = self.knotsZ[self.cents]

        heights = np.full(tuple(np.r_[x.shape, 3]), np.nan)
        ind = self.trifind(x, y)

        for ii in np.unique(ind):
            if ii == -1:
                continue
            indi = ind == ii
            for iref, (i0, i1) in enumerate([(1, 2), (2, 0), (0, 1)]):
                v_base = np.array([
                    R[ii, i1] - R[ii, i0],
                    Z[ii, i1] - Z[ii, i0],
                ])
                v_perp = np.array([v_base[1], -v_base[0]])
                v_base = v_base / np.linalg.norm(v_base)
                v_perp = v_perp / np.linalg.norm(v_perp)

                v0 = np.array([
                    R[ii, i0] - R[ii, iref],
                    Z[ii, i0] - Z[ii, iref],
                ])
                v0_base = v0[0]*v_base[0] + v0[1]*v_base[1]
                v0_perp = v0[0]*v_perp[0] + v0[1]*v_perp[1]

                v_height = (v0 + (-v0_base*v_base + v0_perp*v_perp))/2.
                v_height_norm = np.linalg.norm(v_height)

                dR = x[indi] - R[ii, iref]
                dZ = y[indi] - Z[ii, iref]
                heights[indi, iref] = (
                    dR*v_height[0] + dZ*v_height[1]
                ) / v_height_norm**2

        indok = ~np.isnan(heights)
        assert np.all(heights[indok] >= 0. - 1e-14)
        assert np.all(heights[indok] <= 1. + 1e-14)
        return heights, ind

    # --------
    # bsplines

    def _get_knotscents_per_bs(
        self,
        ind=None,
        return_cents=None,
        return_knots=None,
        returnas=None,
    ):
        """ Return 2 arrays of int indices

        A (nbs, ncents_per_bs) array
        A (nbs, nknots_per_bs) array
        """

        # ------------
        # check inputs

        return_cents = ds._generic_check._check_var(
            return_cents, 'return_cents',
            types=bool,
            default=True,
        )

        return_knots = ds._generic_check._check_var(
            return_knots, 'return_knots',
            types=bool,
            default=True,
        )

        returnas = ds._generic_check._check_var(
            returnas, 'returnas',
            types=str,
            allowed=['ind', 'data'],
            default='ind',
        )

        if ind is None:
            ind = np.ones((self.nbs,), dtype=bool)
        ind_num = ind.nonzero()[0]
        nbs = ind.sum()

        # ------------
        # added for details

        if self.deg == 0:
            if return_cents:
                cents_per_bs = ind_num[:, None]
            if return_knots:
                knots_per_bs = self.cents[ind, :]

        elif self.deg == 1:
            if return_cents or return_knots:
                cents_per_bs = self.cents_per_knots[ind, :]
            if return_knots:
                nmax_cents = np.sum(cents_per_bs >= 0, axis=1)
                nmax = self.cents_per_knots.shape[1] + 3
                knots_per_bs = -np.ones((nbs, nmax), dtype=int)
                knots_per_bs[:, 0] = ind_num
                for ii, i0 in enumerate(ind_num):
                    nu = np.unique(
                        self.cents[cents_per_bs[ii, :nmax_cents[ii]], :]
                    )
                    knots_per_bs[ii, 1:nu.size] = [nn for nn in nu if nn != i0]

        elif self.deg == 2:
            raise NotImplementedError()

        # ------
        # return

        if returnas == 'data':

            # cents
            if return_cents:
                nmax = np.sum(cents_per_bs >= 0, axis=1)
                cents_per_bs_temp = np.full((2, nbs, nmax.max()), np.nan)
                for ii in range(nbs):
                    ind_temp = self.cents[cents_per_bs[ii, :nmax[ii]], :]
                    cents_per_bs_temp[0, ii, :nmax[ii]] = np.mean(
                        self.knotsR[ind_temp],
                        axis=1,
                    )
                    cents_per_bs_temp[1, ii, :nmax[ii]] = np.mean(
                        self.knotsZ[ind_temp],
                        axis=1,
                    )
                cents_per_bs = cents_per_bs_temp

            # knots
            if return_knots:
                nmax = np.sum(knots_per_bs >= 0, axis=1)
                knots_per_bs_temp = np.full((2, nbs, nmax.max()), np.nan)
                for ii in range(nbs):
                    ind_temp = knots_per_bs[ii, :nmax[ii]]
                    knots_per_bs_temp[0, ii, :nmax[ii]] = self.knotsR[ind_temp]
                    knots_per_bs_temp[1, ii, :nmax[ii]] = self.knotsZ[ind_temp]
                knots_per_bs = knots_per_bs_temp

        # return
        if return_cents and return_knots:
            return knots_per_bs, cents_per_bs
        elif return_cents:
            return cents_per_bs
        elif return_knots:
            return knots_per_bs

    def _get_bs_cents(
        self,
        ind=None,
    ):
        """ Return (2, nbs) array of cordinates of centers per bspline

        """

        # ------------
        # check inputs

        if ind is None:
            ind = np.ones((self.nbs,), dtype=bool)

        # ------------
        # added for details

        if self.deg == 0:
            bs_cents = np.array([
                np.mean(self.knotsR[self.cents[ind, :]], axis=1),
                np.mean(self.knotsZ[self.cents[ind, :]], axis=1),
            ])

        elif self.deg == 1:
            bs_cents = np.array([
                self.knotsR[ind],
                self.knotsZ[ind],
            ])

        elif self.deg == 2:
            raise NotImplementedError()

        return bs_cents

    # --------
    # evaluation checks

    def _check_coefs(self, coefs=None):
        """ None for ev_details, (nt, shapebs) for sum """
        if coefs is not None:
            assert coefs.ndim == len(self.shapebs) + 1
            assert coefs.shape[1:] == self.shapebs

    def _ev_generic(
        self,
        x,
        y,
        indbs=None,
    ):
        # -----------
        # check input

        if indbs is None:
            indbs = np.ones((self.nbs,), dtype=bool)
        else:
            indbs = np.atleast_1d(indbs).ravel()

        c0 = (
            isinstance(indbs, np.ndarray)
            and (
                ('bool' in indbs.dtype.name and indbs.size == self.nbs)
                or ('int' in indbs.dtype.name)
            )
        )
        if not c0:
            msg = (
                "Arg indbs must be  a (nbs,) bool or int array!"
                "\nProvided: {indbs}"
            )
            raise Exception(msg)

        if 'bool' in indbs.dtype.name:
            indbs = indbs.nonzero()[0]

        # indbs => indcent : triangles which are ok
        knots_per_bs, cents_per_bs = self._get_knotscents_per_bs(
            return_cents=True,
            return_knots=True,
            returnas='ind',
        )
        indcent = np.unique(cents_per_bs[indbs, :])
        indcent = indcent[indcent >= 0]

        # -----------
        # prepare

        nbs = indbs.size

        return nbs, knots_per_bs, cents_per_bs, indcent

    # --------
    # evaluation

    def ev_details(
        self,
        R=None,
        Z=None,
        # for compatibility (unused)
        indbs_tf=None,
        crop=None,
        cropbs=None,
        coefs=None,
        val_out=None,
    ):

        # -----------
        # generic

        # points
        x, y, crop = _mbr._check_RZ_crop(
            R=R,
            Z=Z,
            crop=crop,
            cropbs=cropbs,
        )

        # parameters
        nbs, knots_per_bs, cents_per_bs, indcent = self._ev_generic(
            x, y, indbs=indbs_tf,
        )

        # -----------
        # compute

        val = np.zeros(tuple(np.r_[x.shape, nbs]))
        heights, ind = self.get_heights_per_centsknots_pts(x, y)

        indu = np.unique(ind[ind >= 0])
        if self.deg == 0:
            if indbs_tf is None:
                for ii in np.intersect1d(indu, indcent):
                    indi = ind == ii
                    val[indi, ii] = 1.
            else:
                for ii in np.intersect1d(indu, indcent):
                    indi = ind == ii
                    ibs = indbs_tf == ii
                    val[indi, ibs] = 1.

        elif self.deg == 1:
            if indbs_tf is None:
                for ii in np.intersect1d(indu, indcent):
                    indi = ind == ii
                    # get bs
                    ibs = np.any(cents_per_bs == ii, axis=1).nonzero()[0]
                    sorter = np.argsort(self.cents[ii, :])
                    inum = sorter[np.searchsorted(
                        self.cents[ii, :],
                        knots_per_bs[ibs, 0],
                        sorter=sorter,
                    )]
                    for jj, jbs in enumerate(ibs):
                        val[indi, jbs] = 1. - heights[indi, inum[jj]]
            else:
                for ii in np.intersect1d(indu, indcent):
                    indi = ind == ii
                    # get bs
                    ibs = np.intersect1d(
                        indbs_tf,
                        np.any(cents_per_bs == ii, axis=1).nonzero()[0],
                    )
                    sorter = np.argsort(self.cents[ii, :])
                    inum = sorter[np.searchsorted(
                        self.cents[ii, :],
                        knots_per_bs[ibs, 0],
                        sorter=sorter,
                    )]
                    for jj, jbs in enumerate(ibs):
                        ij = indbs_tf == jbs
                        val[indi, ij] = 1. - heights[indi, inum[jj]]
        return val

    def ev_sum(
        self,
        R=None,
        Z=None,
        coefs=None,
        val_out=None,
        # for compatibility (unused)
        crop=None,
        cropbs=None,
        indbs_tf=None,
    ):

        # -----------
        # generic

        if val_out is None:
            val_out = np.nan

        # coefs
        self._check_coefs(coefs=coefs)

        # points
        x, y, crop = _mbr._check_RZ_crop(
            R=R,
            Z=Z,
            crop=crop,
            cropbs=cropbs,
        )
        nt = coefs.shape[0]

        # parameters
        nbs, knots_per_bs, cents_per_bs, indcent = self._ev_generic(
            x, y, indbs=None,
        )

        # -----------
        # prepare

        val = np.zeros(np.r_[nt, x.shape])
        heights, ind = self.get_heights_per_centsknots_pts(x, y)
        if not np.isscalar(coefs):
            newshape = np.r_[coefs.shape, 1]
            coefs = coefs.reshape(newshape)

        # -----------
        # compute

        indu = np.unique(ind[ind >= 0])
        if self.deg == 0:
            for ii in np.intersect1d(indu, indcent):
                indi = ind == ii
                val[:, indi] += coefs[:, ii, ...]

        elif self.deg == 1:
            for ii in np.intersect1d(indu, indcent):
                indi = ind == ii
                # get bs
                ibs = np.any(cents_per_bs == ii, axis=1).nonzero()[0]
                sorter = np.argsort(self.cents[ii, :])
                inum = sorter[np.searchsorted(
                    self.cents[ii, :],
                    knots_per_bs[ibs, 0],
                    sorter=sorter,
                )]
                for jj, jbs in enumerate(ibs):
                    val[:, indi] += (
                        1. - heights[indi, inum[jj]]
                    ) * coefs[:, jbs, ...]

        if val_out is not False:
            val[:, ind == -1] = val_out
        return val

    # TBC
    def get_overlap(self):
        raise NotImplementedError()
        return _get_overlap(
            deg=self.degrees[0],
            knotsx=self.knots_per_bs_x,
            knotsy=self.knots_per_bs_y,
            shapebs=self.shapebs,
        )

    # TBD / TBF
    def get_operator(
        self,
        operator=None,
        geometry=None,
        cropbs_flat=None,
        # specific to deg = 0
        cropbs=None,
    ):
        """ Get desired operator """
        raise NotImplementedError()
        return _class1_bsplines_operators_tri.get_mesh2dRect_operators(
            deg=self.degrees[0],
            operator=operator,
            geometry=geometry,
            knotsx_mult=self.tck[0],
            knotsy_mult=self.tck[1],
            knotsx_per_bs=self.knots_per_bs_x,
            knotsy_per_bs=self.knots_per_bs_y,
            overlap=self.get_overlap(),
            cropbs_flat=cropbs_flat,
            cropbs=cropbs,
        )


# #############################################################################
# #############################################################################
#                       Mesh2Dtri - bsplines - overlap
# #############################################################################


def _get_overlap(
    deg=None,
    knotsx=None,
    knotsy=None,
    shapebs=None,
):
    raise NotImplementedError()
    return inter


# #############################################################################
# #############################################################################
#                           Mesh2DRect - bsplines
# #############################################################################


def get_bs2d_func(
    knotsR=None,
    knotsZ=None,
    cents=None,
    trifind=None,
    deg=None,
):

    # -----------------
    # Instanciate class

    clas = BivariateSplineTri(
        knotsR=knotsR,
        knotsZ=knotsZ,
        cents=cents,
        trifind=trifind,
        deg=deg,
    )

    return clas.ev_details, clas.ev_sum, clas



# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# Also if needed: retab
'''
    Magnetic ripple
'''

import warnings
import traceback

try:
    try:
        from tofu.mag.mag_ripple import *
    except Exception:
        from .mag_ripple import *
    del warnings, traceback
except Exception as err:
    msg = str(traceback.format_exc())
    msg += "\n\n    => the optional sub-package is not usable\n"
    warnings.warn(msg)
    del msg, err

__all__ = ['mag_ripple']





import os
import sys
import warnings
import copy


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import datastock as ds


from . import _spectrallines_checks
from . import _spectrallines_compute
from . import _spectrallines_plot


__all__ = ['SpectralLines']


_WHICH_LINES = 'lines'
_QUANT_NE = 'ne'
_QUANT_TE = 'Te'
_UNITS_LAMBDA0 = 'm'


#############################################
#############################################
#       Spectral Lines
#############################################


class SpectralLines(ds.DataStock):

    _ddef = copy.deepcopy(ds.DataStock._ddef)
    _ddef['params']['dobj'] = {
        'lines': {
            'lambda0': (float, 0.),
            'source': (str, 'unknown'),
            'transition':    (str, 'unknown'),
            'element':  (str, 'unknown'),
            'charge':  (int, 0),
            'ion':  (str, 'unknown'),
            'symbol':   (str, 'unknown'),
        },
    }

    _show_in_summary_core = ['shape', 'ref']
    _show_in_summary = 'all'

    _which_lines = _WHICH_LINES
    _quant_ne = _QUANT_NE
    _quant_Te = _QUANT_TE

    _units_lambda0 = _UNITS_LAMBDA0

    def add_line(
        self,
        key=None,
        lambda0=None,
        pec=None,
        source=None,
        transition=None,
        ion=None,
        symbol=None,
        **kwdargs,
    ):
        """ Add a spectral line by key and rest wavelength, optionally with

        """
        self.add_obj(
            which='lines',
            key=key,
            lambda0=lambda0,
            pec=pec,
            source=source,
            transition=transition,
            ion=ion,
            symbol=symbol,
            **kwdargs,
        )

    def add_pec(self, key=None, pec=None, ref=None):
        pass

    # -----------------
    # from openadas
    # ------------------

    @classmethod
    def from_openadas(
        cls,
        lambmin=None,
        lambmax=None,
        element=None,
        charge=None,
        online=None,
        update=None,
        create_custom=None,
    ):
        """
        Load lines and pec from openadas, either:
            - online = True:  directly from the website
            - online = False: from pre-downloaded files in ~/.tofu/openadas/
        """
        ddata, dref, dobj = _spectrallines_compute.from_openadas(
            lambmin=lambmin,
            lambmax=lambmax,
            element=element,
            charge=charge,
            online=online,
            update=update,
            create_custom=create_custom,
            which_lines=cls._which_lines,
        )
        return cls(ddata=ddata, dref=dref, dobj=dobj)

    def add_from_openadas(
        self,
        lambmin=None,
        lambmax=None,
        element=None,
        charge=None,
        online=None,
        update=None,
        create_custom=None,
    ):
        """
        Load and add lines and pec from openadas, either:
            - online = True:  directly from the website
            - online = False: from pre-downloaded files in ~/.tofu/openadas/
        """
        ddata, dref, dobj = _spectrallines_compute.from_openadas(
            lambmin=lambmin,
            lambmax=lambmax,
            element=element,
            charge=charge,
            online=online,
            update=update,
            create_custom=create_custom,
            dsource0=self._dobj.get('source'),
            dref0=self._dref,
            ddata0=self._ddata,
            dobj0=self._dobj,
            which_lines=self._which_lines,
        )
        self.update(ddata=ddata, dref=dref, dobj=dobj)

    # -----------------
    # from nist
    # ------------------

    @classmethod
    def from_nist(
        cls,
        lambmin=None,
        lambmax=None,
        element=None,
        charge=None,
        ion=None,
        wav_observed=None,
        wav_calculated=None,
        transitions_allowed=None,
        transitions_forbidden=None,
        cache_from=None,
        cache_info=None,
        verb=None,
        create_custom=None,
    ):
        """
        Load lines and pec from openadas, either:
            - online = True:  directly from the website
            - online = False: from pre-downloaded files in ~/.tofu/openadas/
        """
        dobj = _spectrallines_compute._from_nist(
            lambmin=lambmin,
            lambmax=lambmax,
            element=element,
            charge=charge,
            ion=ion,
            wav_observed=wav_observed,
            wav_calculated=wav_calculated,
            transitions_allowed=transitions_allowed,
            transitions_forbidden=transitions_forbidden,
            cache_from=cache_from,
            cache_info=cache_info,
            verb=verb,
            create_custom=create_custom,
            group_lines=cls._which_lines,
        )
        return cls(dobj=dobj)

    def add_from_nist(
        self,
        lambmin=None,
        lambmax=None,
        element=None,
        charge=None,
        ion=None,
        wav_observed=None,
        wav_calculated=None,
        transitions_allowed=None,
        transitions_forbidden=None,
        cache_from=None,
        cache_info=None,
        verb=None,
        create_custom=None,
    ):
        """
        Load and add lines and pec from openadas, either:
            - online = True:  directly from the website
            - online = False: from pre-downloaded files in ~/.tofu/openadas/
        """
        dobj = _spectrallines_compute._from_nist(
            lambmin=lambmin,
            lambmax=lambmax,
            element=element,
            charge=charge,
            ion=ion,
            wav_observed=wav_observed,
            wav_calculated=wav_calculated,
            transitions_allowed=transitions_allowed,
            transitions_forbidden=transitions_forbidden,
            cache_from=cache_from,
            cache_info=cache_info,
            verb=verb,
            create_custom=create_custom,
            dsource0=self._dobj.get('source'),
            dlines0=self._dobj.get('lines'),
            group_lines=self._which_lines,
        )
        self.update(dobj=dobj)

    # -----------------
    # from file (.py)
    # ------------------

    @classmethod
    def from_module(cls, pfe=None):

        dobj = _spectrallines_compute.from_module(pfe=pfe)

        # Create collection
        out = cls(dobj=dobj)

        # Replace ION by ion if relevant
        c0 = (
            'ion' in out._dobj.keys()
            and 'ion' not in out.get_lparam(which='lines')
            and 'ION' in out.get_lparam(which='lines')
        )
        if c0:
            for k0, v0 in out._dobj['lines'].items():
                ion = [
                    k1 for k1, v1 in out._dobj['ion'].items()
                    if out._dobj['lines'][k0]['ION'] == v1['ION']
                ][0]
                out._dobj['lines'][k0]['ion'] = ion
                del out._dobj['lines'][k0]['ION']
        return out

    # -----------------
    # summary
    # ------------------

    # -----------------
    # conversion wavelength - energy - frequency
    # ------------------

    def convert_lines(self, units=None, key=None, ind=None, returnas=None):
        """ Convert wavelength (m) to other units or other quantities

        Avalaible units:
            wavelength: km, m, mm, um, nm, pm, A
            energy:     J, eV, keV, MeV, GeV
            frequency:  Hz, kHz, MHz, GHz, THz

        Return the result as a np.ndarray (returnas = 'data')

        Can also just return the conversion coef if returnas='coef'
        In that case, a bool is also returned indicating whether the result is
        the proportional to the inverse of lambda0::
            - False: data = coef * lambda0
            - True: data = coef / lambda0
        """
        if units is None:
            units = self._units_lambda0
        if returnas is None:
            returnas = dict

        lok = [dict, np.ndarray, 'data', 'coef']
        if returnas not in lok:
            msg = (
                "Arg returnas must be in:\n"
                + "\t- {}\n".format(lok)
                + "\t- provided: {}".format(returnas)
            )
            raise Exception(msg)
        if returnas in [dict, np.ndarray, 'data']:
            returnas2 = 'data'
        else:
            returnas2 = 'coef'

        # get keys of desired lines
        key = self._ind_tofrom_key(
            which=self._which_lines, key=key, ind=ind, returnas=str,
        )

        # get wavelength in m
        lamb_in = self.get_param(
            which=self._which_lines, param='lambda0',
            key=key, returnas=np.ndarray,
        )['lambda0']

        # conversion
        out = _spectrallines_compute.convert_spectral(
            data_in=lamb_in,
            units_in='m',
            units_out=units,
            returnas=returnas2,
        )
        if returnas is dict:
            out = {k0: out[ii] for ii, k0 in enumerate(key)}
        return out

    # -----------------
    # PEC interpolation
    # ------------------

    def calc_pec(
        self,
        key=None,
        ind=None,
        ne=None,
        Te=None,
        deg=None,
        grid=None,
        return_params=None,
    ):
        """ Compute the pec (<sigma v>) by interpolation for chosen lines

        Assumes Maxwellian electron distribution

        Provide ne and Te and 1d np.ndarrays

        if grid=False:
            - ne is a (n,) 1d array
            - Te is a (n,) 1d array
          => the result is a dict of (n,) 1d array

        if grid=True:
            - ne is a (n1,) 1d array
            - Te is a (n2,) 1d array
          => the result is a dict of (n1, n2) 2d array
        """

        # Check keys
        key = self._ind_tofrom_key(
            which=self._which_lines, key=key, ind=ind, returnas=str,
        )
        dlines = self._dobj[self._which_lines]

        key, dnTe, return_params = _spectrallines_checks._check_compute_pec(
            # check keys
            key=key,
            dlines=dlines,
            ddata=self._ddata,
            _quant_ne=self._quant_ne,
            _quant_Te=self._quant_Te,
            # check ne, Te
            ne=ne,
            Te=Te,
            return_params=return_params,
        )
        keypec = [f'{k0}-pec' for k0 in key]

        # group lines per common ref
        lref = set([self._ddata[k0]['ref'] for k0 in keypec])
        dref = {
            k0: [k1 for k1 in keypec if self._ddata[k1]['ref'] == k0]
            for k0 in lref
        }

        # Interpolate
        for ii, (k0, v0) in enumerate(dref.items()):
            douti, dparami = self.interpolate(
                # interpolation base
                keys=v0,
                ref_keys=None,
                ref_quant=[self._quant_ne, self._quant_Te],
                # interpolation pts
                pts_axis0=dnTe['ne'],
                pts_axis1=dnTe['Te'],
                # parameters
                deg=deg,
                deriv=0,
                grid=grid,
                log_log=True,
                return_params=True,
            )

            # update dict
            if ii == 0:
                dout = douti
                dparam = dparami
            else:
                dout.update(**douti)
                dparam['keys'] += dparami['keys']
                dparam['ref_keys'] += dparami['ref_keys']

        # -------
        # return

        if return_params is True:
            dparam['key'] = dparam['keys']
            del dparam['keys']
            dparam['ne'] = dparam['pts_axis0']
            dparam['Te'] = dparam['pts_axis1']
            del dparam['pts_axis0'], dparam['pts_axis1'], dparam['pts_axis2']
            return dout, dparam
        else:
            return dout

    def calc_intensity(
        self,
        key=None,
        ind=None,
        ne=None,
        Te=None,
        concentration=None,
        deg=None,
        grid=None,
    ):
        """ Compute the lines intensities by pec interpolation for chosen lines

        Assumes Maxwellian electron distribution
        Assumes concentration = nz / ne

        Provide ne and Te and 1d np.ndarrays

        Provide concentration as:
            - a np.ndarray (same concentration assumed for all lines)
            - a dict of {key: np.ndarray}

        if grid=False:
            - ne is a (n,) 1d array
            - Te is a (n,) 1d array
            - concentration is a (dict of) (n,) 1d array(s)
          => the result is a dict of (n1, n2) 2d array

        if grid=True:
            - ne is a (n1,) 1d array
            - Te is a (n2,) 1d array
            - concentration is a (dict of) (n1, n2) 2d array(s)
          => the result is a dict of (n1, n2) 2d array


        """

        # Check keys
        key = self._ind_tofrom_key(
            which=self._which_lines, key=key, ind=ind, returnas=str,
        )

        # interpolate pec
        dout, dparam = self.calc_pec(
            key=key,
            ind=ind,
            ne=ne,
            Te=Te,
            grid=grid,
            deg=deg,
            return_params=True,
        )

        # check concentrations
        concentration = _spectrallines_checks._check_compute_intensity(
            key=[k0[:-4] for k0 in dparam['key']],
            concentration=concentration,
            shape=dparam['ne'].shape,
        )

        # Derive intensity
        for k0, v0 in dout.items():
            dout[k0] = v0*dparam['ne']**2*concentration[k0[:-4]]

        return dout

    # -----------------
    # plotting
    # ------------------

    def plot_spectral_lines(
        self,
        key=None,
        ind=None,
        ax=None,
        sortby=None,
        param_x=None,
        param_txt=None,
        ymin=None,
        ymax=None,
        ls=None,
        lw=None,
        fontsize=None,
        side=None,
        dsize=None,
        dcolor=None,
        fraction=None,
        figsize=None,
        dmargin=None,
        wintit=None,
        tit=None,
    ):
        """ plot rest wavelengths as vertical lines """

        # Check inputs
        key = self._ind_tofrom_key(
            which=self._which_lines, key=key, ind=ind, returnas=str,
        )

        return _spectrallines_plot.plot_axvlines(
            din=self._dobj[self._which_lines],
            key=key,
            param_x=param_x,
            param_txt=param_txt,
            sortby=sortby,
            dsize=dsize,
            ax=ax, ymin=ymin, ymax=ymax,
            ls=ls, lw=lw, fontsize=fontsize,
            side=side, dcolor=dcolor, fraction=fraction,
            figsize=figsize, dmargin=dmargin,
            wintit=wintit, tit=tit,
        )

    def plot_pec_single(
        self,
        key=None,
        ind=None,
        ne=None,
        Te=None,
        concentration=None,
        deg=None,
        grid=None,
        ax=None,
        sortby=None,
        param_x=None,
        param_txt=None,
        ymin=None,
        ymax=None,
        ls=None,
        lw=None,
        fontsize=None,
        side=None,
        dcolor=None,
        fraction=None,
        figsize=None,
        dmargin=None,
        wintit=None,
        tit=None,
    ):
        """ Same as plot_spectral_lines() with extra scatter plot with circles

        The circles' diameters depend on the pec value for each line

        Requires:
            - Te = scalar (eV)
            - ne = scalar (/m3)

        """

        # ------------
        # Check ne, Te

        ltypes = [int, float, np.integer, np.floating]
        dnTe = {'ne': ne, 'Te': Te}
        single = all([
            type(v0) in ltypes or len(v0) == 1 for v0 in dnTe.values()
        ])
        if not single:
            msg = ("Arg ne and Te must be floats!")
            raise Exception(msg)

        # --------
        # Get dpec
        dpec = self.calc_pec(
            key=key,
            ind=ind,
            ne=ne,
            Te=Te,
            deg=deg,
            grid=False,
            return_params=False,
        )
        key = [k0[:-4] for k0 in dpec.keys()]

        ne = float(ne)
        Te = float(Te)
        tit = (
            r'$n_e$' + f'= {ne} ' + r'$/m^3$'
            + r' -  $T_e$ = ' + f'{Te/1000.} keV'
        )

        pmax = np.max([np.log10(v0) for v0 in dpec.values()])
        pmin = np.min([np.log10(v0) for v0 in dpec.values()])
        dsize = {
            k0[:-4]: (np.log10(v0) - pmin) / (pmax - pmin)*19 + 1
            for k0, v0 in dpec.items()
        }

        return _spectrallines_plot.plot_axvlines(
            din=self._dobj[self._which_lines],
            key=key,
            param_x=param_x,
            param_txt=param_txt,
            sortby=sortby,
            dsize=dsize,
            ax=ax, ymin=ymin, ymax=ymax,
            ls=ls, lw=lw, fontsize=fontsize,
            side=side, dcolor=dcolor, fraction=fraction,
            figsize=figsize, dmargin=dmargin,
            wintit=wintit, tit=tit,
        )

    def plot_pec(
        self,
        key=None,
        ind=None,
        ne=None,
        Te=None,
        norder=None,
        ne_scale=None,
        Te_scale=None,
        param_txt=None,
        param_color=None,
        deg=None,
        dax=None,
        proj=None,
        ymin=None,
        ymax=None,
        ls=None,
        lw=None,
        fontsize=None,
        side=None,
        dcolor=None,
        fraction=None,
        figsize=None,
        dmargin=None,
        dtit=None,
        tit=None,
        wintit=None,
    ):

        # -----------
        # Check input

        # Check ne, Te
        lc = [np.isscalar(ne) or len(ne) == 1, np.isscalar(Te) or len(Te) == 1]
        if all(lc):
            msg = "For a single (ne, Te) space, use plot_pec_singe()"
            raise Exception(msg)

        # Get dpec
        dpec = self.calc_pec(
            key=key,
            ind=ind,
            ne=ne,
            Te=Te,
            deg=deg,
            grid=False,
        )
        damp = {k0: {'data': v0} for k0, v0 in dpec.items()}

        # Create grid
        ne_grid = ds._class1_compute._get_grid1d(
            ne, scale=ne_scale, npts=ne.size*2, nptsmin=3,
        )
        Te_grid = ds._class1_compute._get_grid1d(
            Te, scale=Te_scale, npts=Te.size*2, nptsmin=3,
        )

        # get dpec for grid
        dpec_grid = self.calc_pec(
            key=key,
            ind=ind,
            ne=ne_grid,
            Te=Te_grid,
            deg=deg,
            grid=True,
        )

        raise NotImplementedError()

        return _spectrallines_plot.plot_dominance_map(
            din=self._dobj['lines'], im=im, extent=extent,
            xval=ne, yval=Te, damp=damp,
            x_scale=ne_scale, y_scale=Te_scale, amp_scale='log',
            param_txt='symbol',
            dcolor=dcolor,
            dax=dax, proj=proj,
            figsize=figsize, dmargin=dmargin,
            wintit=wintit, tit=tit, dtit=dtit,
        )





# Standard
import itertools as itt
import warnings


# Common
import numpy as np
import scipy.sparse as scpsp
import datastock as ds


_DRESERVED_KEYS = {
    'dref': ['ldata', 'ldata_monot', 'size', 'ind'],
    'dstatic': [],
    'ddata': ['ref', 'shape', 'data'],
    'dobj': [],
}
_LRESERVED_KEYS = list(set(itt.chain.from_iterable([
    v0 for v0 in _DRESERVED_KEYS.values()
])))


_DDEF_PARAMS = {
    'ddata': {
        'source': (str, 'unknown'),
        'dim':    (str, 'unknown'),
        'quant':  (str, 'unknown'),
        'name':   (str, 'unknown'),
        'units':  (str, 'a.u.'),
    },
    'dobj': {
    },
    'dstatic': {
    },
}


_IREF = 'iref'
_IDATA = 'data'


_DATA_NONE = False


# #############################################################################
# #############################################################################
#                           Generic
# #############################################################################


def _check_which(
    dref=None,
    ddata=None,
    dobj=None,
    dstatic=None,
    which=None,
    return_dict=None,
):
    """ Check which in ['ref', 'data'] + dobj.keys() + dstatic.keys()

    Optionally return the dict itself, by reference

    """

    # --------------
    # Check inputs

    return_dict = ds._generic_check._check_var(
        return_dict,
        'return_dict',
        types=bool,
        default=True,
    )

    lkobj = list(dobj.keys())
    lkstatic = list(dstatic.keys())
    lkok = ['ref', 'data'] + lkobj + lkstatic
    which = ds._generic_check._check_var(
        which,
        'which',
        types=str,
        allowed=lkok,
        default='data',
    )

    # -----------------
    # return right dict

    if return_dict is True:
        if which == 'ref':
            dd = dref
        elif which == 'data':
            dd = ddata
        elif which in lkobj:
            dd = dobj[which]
        else:
            dd = dstatic[which]
        return which, dd
    else:
        return which


def _check_conflicts(dd=None, dd0=None, dd_name=None, returnas=None):
    """ Detect conflict with existing entries

    Any pre-existing entry will trigger either an update or a conflict

    - conflic: same parameter, different value
    - update: new parameter
    - retro: parameter not filled (None)

    """

    # ------------
    # check inputs

    if returnas is None:
        returnas = False

    # ----------------------------
    # detect conflicts and updates

    dupdate = {}
    dconflict = {}
    for k0, v0 in dd.items():

        # k0 not in existing dict => ok
        if k0 not in dd0.keys():
            continue

        # find conflicts (same key and same parameters with different values)
        lk = set(v0.keys()).intersection(dd0[k0].keys())
        lk = [
            kk for kk in lk
            if not (
                isinstance(v0[kk], dd0[k0][kk].__class__)
                and (
                    (
                        isinstance(v0[kk], np.ndarray)
                        and v0[kk].shape == dd0[k0][kk].shape
                        and np.allclose(v0[kk], dd0[k0][kk], equal_nan=True)
                    )
                    or (
                        scpsp.issparse(v0[kk])
                        and v0[kk].shape == dd0[k0][kk].shape
                        and np.allclose(
                            v0[kk].data,
                            dd0[k0][kk].data,
                            equal_nan=True,
                        )
                    )
                    or (
                        not isinstance(v0[kk], np.ndarray)
                        and not scpsp.issparse(v0[kk])
                        and v0[kk] == dd0[k0][kk]
                    )
                    or (
                        v0[kk] == dd0[k0][kk]
                    )
                )
            )
        ]
        if len(lk) > 0:
            dconflict[k0] = lk

        # find updates (same key but new parameters)
        lkup = [
            kk for kk in v0.keys()
            if kk not in lk
            and kk not in dd0[k0].keys()
            # and kk not in ['ldata', 'size']
            and kk not in _LRESERVED_KEYS
        ]
        if len(lkup) > 0:
            dupdate[k0] = lkup

    # ---------------
    # raise or return

    if returnas is False:
        # Conflicts => Exception
        if len(dconflict) > 0:
            lstr = [
                f"\t- {dd_name}['{k0}']: {v0}"
                for k0, v0 in dconflict.items()
            ]
            msg = (
                f"Conflicts with pre-existing values found in {dd_name}:\n"
                + "\n".join(lstr)
            )
            raise Exception(msg)

        # Updates => Warning
        if len(dupdate) > 0:
            lstr = [
                f"\t- {dd_name}['{k0}']: {v0}"
                for k0, v0 in dupdate.items()
            ]
            msg = (
                f"\nExisting {dd_name} keys updated with new keys:\n"
                + "\n".join(lstr)
            )
            warnings.warn(msg)

    else:
        return dconflict, dupdate


def _check_remove(key=None, dkey=None, name=None):
    c0 = isinstance(key, str) and key in dkey.keys()
    c1 = (
        isinstance(key, list)
        and all([isinstance(kk, str) and kk in dkey.keys() for kk in key])
    )
    if not (c0 or c1):
        msg = (
            """
            Removed param must be a str already in self.d{}
            It can also be a list of such
            \t- provided: '{}'
            \t- already available: {}
            """.format(name, key, sorted(dkey.keys()))
        )
        raise Exception(msg)
    if c0:
        key = [key]
    return key


# #############################################################################
# #############################################################################
#                           Removing routines
# #############################################################################


def _get_whichorkey(key=None, which=None, din=None, dname=None):

    # -------------
    # trivial check

    lc = [
        key is not None,
        which is not None,
    ]
    if np.sum(lc) == 0:
        msg = "Please provide either key or/xor which!"
        raise Exception(msg)

    # ----------
    # which only

    if lc[1]:
        which = ds._generic_check._check_var(
            which, 'which',
            allowed=sorted(din.keys()),
        )

    # ---------------------------------------
    # key but not which => check and set which

    elif lc[0] and not lc[1]:

        if hasattr(key, '__iter__'):
            lwhich = [
                k0 for k0, v0 in din.items()
                if key in v0.keys()
            ]
        else:
            lwhich = [
                k0 for k0, v0 in din.items()
                if all([kk in v0.keys() for kk in key])
            ]

        if len(lwhich) == 1:
            which = lwhich[0]

        else:
            msg = (
                f"key {key} has no / several matches in {dname}:\n"
                f"\t- matches: {lwhich}"
            )
            raise Exception(msg)

    if key is not None:
        key = ds._generic_check._check_var_iter(
            key, 'key',
            types=list,
            allowed=sorted(din[which]),
        )

    return key, which


def _remove_ref(
    # key to remove
    key=None,
    # dict
    dref0=None,
    ddata0=None,
    dstatic0=None,
    dobj0=None,
    # parameters
    propagate=None,
    reserved_keys=None,
    ddefparams_data=None,
    ddefparams_obj=None,
    ddefparams_static=None,
    data_none=None,
    max_ndim=None,
):
    """ Remove a ref (or list of refs) and all associated data """

    # trivial case
    if key is None:
        return dref0, dstatic0, ddata0, dobj0

    # check input
    key = _check_remove(
        key=key, dkey=dref0, name='ref',
    )

    for k0 in key:
        # Remove orphan ddata
        for k1 in dref0[k0]['ldata']:
            del ddata0[k1]
        del dref0[k0]

    # Double-check consistency
    return _consistency(
        ddata=None, ddata0=ddata0,
        dref=None, dref0=dref0,
        dstatic=None, dstatic0=dstatic0,
        dobj=None, dobj0=dobj0,
        reserved_keys=reserved_keys,
        ddefparams_data=ddefparams_data,
        ddefparams_obj=ddefparams_obj,
        ddefparams_static=ddefparams_static,
        data_none=data_none,
        max_ndim=max_ndim,
    )


def _remove_static(
    # key to remove
    key=None,
    which=None,
    # dict
    dstatic0=None,
    ddata0=None,
    dobj0=None,
    # parameters
    propagate=None,
):
    """ Remove a static ref (or list) or a whole category

    key os provided:
        => remove only the desired key(s)
            works only if key is not used in ddata and dobj

    which is provided:
        => treated as param, the whole category of ref_static is removed
            if propagate, the parameter is removed from ddata and dobj
    """

    # ------------
    # check inputs

    key, which = _get_whichorkey(
        key=key,
        which=which,
        din=dstatic,
        dname='dstatic',
    )

    # --------------------------------
    # key is None => delete whole dict

    if key is None:

        # Propagate (delete as param in ddata and dobj)
        if propagate is None:
            propagate = True

        if propagate is True:
            # ddata
            if which in list(ddata0.values())[0].keys():
                _remove_param(dd=ddata0, dd_name='ddata', param=which)

            # dobj0
            for k0 in dobj0.keys():
                if which in list(dobj0[k0].values())[0].keys():
                    _remove_param(
                        dd=dobj0[k0],
                        dd_name="ddobj['{}']".format(k0),
                        param=which,
                    )

        # remove
        del dstatic0[which]

    # -------------
    # key and which

    else:

        key = _check_remove(
            key=key,
            dkey=dstatic0[which],
            name=f'static[{k0}]',
        )

        # Make sure key is not used (condition for removing)
        for kk in key:
            lk1 = [
                k1 for k1, v1 in ddata0.items()
                if kk == v1.get(which)
            ]
            lk2 = [
                k1 for k1, v1 in dobj0.items()
                if any([kk == v2.get(which) for v2 in v1.values()])
            ]
            if len(lk1) > 0 or len(lk2) > 0:
                msg = (
                    f"Provided static key ({kk}) is used in:\n"
                    + "\n".join(
                        [f"\t- self.ddata['{k1}']" for k1 in lk1]
                        + [
                            f"\t- self.dobj['{k2}']['{which}']"
                            for k2 in lk2
                        ]
                    )
                )
                raise Exception(msg)
            del dstatic0[k0][kk]


def _remove_data(
    # key to remove
    key=None,
    # dict
    dref0=None,
    ddata0=None,
    dstatic0=None,
    dobj0=None,
    # parameters
    propagate=None,
    reserved_keys=None,
    ddefparams_data=None,
    ddefparams_obj=None,
    ddefparams_static=None,
    data_none=None,
    max_ndim=None,
):
    """ Remove a ref (or list of refs) and all associated data """

    # ------------
    # trivial case

    if key is None:
        return dref0, dstatic0, ddata0, dobj0

    # ------------
    # check inputs

    key = _check_remove(
        key=key, dkey=ddata0, name='data',
    )

    # ------
    # remove

    for k0 in key:
        # Remove key from dref['ldata']
        for k1 in dref0.keys():
            if k0 in dref0[k1]['ldata']:
                dref0[k1]['ldata'].remove(k0)
        del ddata0[k0]

    # Propagate upward to ref
    if propagate is True:
        lk = [
            k0 for k0, v0 in dref0.items()
            if len(dref0[k0].get('ldata', [])) == 0
        ]
        for kk in lk:
            del dref0[kk]

    # Double-check consistency
    return _consistency(
        ddata=None, ddata0=ddata0,
        dref=None, dref0=dref0,
        dstatic=None, dstatic0=dstatic0,
        dobj=None, dobj0=dobj0,
        reserved_keys=reserved_keys,
        ddefparams_data=ddefparams_data,
        ddefparams_obj=ddefparams_obj,
        ddefparams_static=ddefparams_static,
        data_none=data_none,
        max_ndim=max_ndim,
    )


def _remove_obj(
    # key to remove
    key=None,
    which=None,
    # dict
    dobj0=None,
    ddata0=None,
    dref0=None,
    dstatic0=None,
    # parameters
    reserved_keys=None,
    ddefparams_data=None,
    ddefparams_obj=None,
    ddefparams_static=None,
    data_none=None,
    max_ndim=None,
):

    # ------------
    # check inputs

    key, which = _get_whichorkey(
        key=key,
        which=which,
        din=dobj,
        dname='dobj',
    )

    # --------------------------------
    # key is None => delete whole dict

    if key is None:

        # remove
        del dstatic0[which]

    # ------------
    # Check inputs

    if key is not None:
        key = _check_remove(
            key=key,
            dkey=dobj0[which],
            name=f"dobj['{which}']",
        )
        for kk in set(key).intersection(dobj0[which].keys()):
            del dobj0[k0][kk]

    return _consistency(
        ddata=None, ddata0=ddata0,
        dref=None, dref0=dref0,
        dstatic=None, dstatic0=dstatic0,
        dobj=None, dobj0=dobj0,
        reserved_keys=reserved_keys,
        ddefparams_data=ddefparams_data,
        ddefparams_obj=ddefparams_obj,
        ddefparams_static=ddefparams_static,
        data_none=data_none,
        max_ndim=max_ndim,
    )


# #############################################################################
# #############################################################################
#                           dstatic
# #############################################################################


def _check_dstatic(
    dstatic=None, dstatic0=None,
):
    """ Check and format dstatic

    dstatic can be:
        - dict

    """

    # ----------------
    # Trivial case
    if dstatic in [None, {}]:
        return {}

    # ----------------
    # Check conformity

    c0 = (
        isinstance(dstatic, dict)
        and all([
            isinstance(k0, str)
            and isinstance(v0, dict)
            and all([
                isinstance(k1, str)
                and isinstance(v1, dict)
                for k1, v1 in v0.items()
            ])
            for k0, v0 in dstatic.items()
        ])
    )

    # Raise exception if non-conformity
    if not c0:
        msg = (
            "Arg dstatic must be a dict of the form:\n"
            "\t dict(\n"
            "\t     'type0': {\n"
            "\t         'k0': {...},\n"
            "\t         'k1': {...}\n"
            "\t     },\n"
            "\t     'type1': {\n"
            "\t         'k2': {...},\n"
            "\t         'k3': {...}\n"
            "\t     },\n"
            "\t )\n"
            f"\nProvided:\n{dstatic}"
        )
        raise Exception(msg)

    # ----------------------------
    # Identify conflicts / updates

    # raise except if conflict with existing entry
    dupdate = {}
    dconflict = {}
    for k0, v0 in dstatic.items():

        if k0 not in dstatic0.keys():
            continue

        dconflict0, dupdate0 = _check_conflicts(
            dd=dstatic[k0],
            dd0=dstatic0[k0],
            dd_name=f"dstatic['{k0}']",
            returnas=True,
        )

        if len(dconflict0) > 0:
            dconflict[k0] = {k1: list(v1) for k1, v1 in dconflict0.items()}

        if len(dupdate0) > 0:
            dupdate[k0] = {k1: list(v1) for k1, v1 in dupdate0.items()}

    # Conflicts => Exception
    if len(dconflict) > 0:
        msg = (
            "The following dstatic keys are conflicting existing values:\n"
            + "\n".join([
                "\t- dstatic['{}']['{}']: {}".format(k0, v0[0], v0[1])
                for k0, v0 in dconflict.items()
            ])
        )
        raise Exception(msg)

    # Updates => Warning
    if len(dupdate) > 0:
        lstr = list(itt.chain.from_iterable([
            [
                f"\t- dstatic['{k0}']['{k1}']: {v1}"
                for k1, v1 in v0.items()
            ]
            for k0, v0 in dupdate.items()
        ]))
        msg = (
            "\nThe following keys will be added to dstatic:\n"
            + "\n".join(lstr)
        )
        warnings.warn(msg)

    # ------------------
    # Check element / ion / charge

    # _check_elementioncharge_dict(dstatic=dstatic)

    return dstatic


# #############################################################################
# #############################################################################
#                           dref
# #############################################################################


class DataRefException(Exception):

    def __init__(self, ref=None, data=None):
        msg = (
            """
            To be a valid reference for {}, provided data must be either:
            \t- np.ndarray:  of dimension 1 with increasing values
            \t- list, tuple: convertible to the above
            \t- dict / other class: used for meshes

            You provided:
            \t- {}

            """.format(ref, data)
        )
        self.message = msg


def _check_dataref(data=None, key=None):
    """ Check the conformity of data to be a valid reference """

    # if not array
    # => try converting or get class (dict, mesh...)
    if not isinstance(data, np.ndarray):
        if isinstance(data, list) or isinstance(data, tuple):
            try:
                data = np.array(data)
                size = data.size
            except Exception as err:
                raise DataRefException(ref=key, data=data)
        else:
            try:
                data, size = _check_mesh_temp(data=data, key=key)
                if len(size) == 1:
                    size = size[0]
            except Exception as err:
                size = data.__class__.__name__

    # if array => check unique (unique + sorted)
    if isinstance(data, np.ndarray):
        if data.ndim != 1:
            raise DataRefException(ref=key, data=data)

        datau = np.unique(data)
        if not (datau.size == data.size and np.allclose(datau, data)):
            raise DataRefException(ref=key, data=data)
        size = data.size

    return data, size


def _check_dref(
    dref=None,
    dref0=None,
    ddata0=None,
):
    """ Check and format dref

    dref can be:
        - dict

    If some data is provided
        => returns ddata to be added

    """

    # ----------------
    # Trivial case
    if dref in [None, {}]:
        return {}, {}

    # ----------------
    # Check conformity

    # Basis
    # lk_opt = ['ldata', 'size', 'data']
    if not isinstance(dref, dict):
        msg = "Arg dref must be a dict!"
        raise Exception(msg)

    dref2 = {}
    for k0, v0 in dref.items():

        # key
        key = ds._generic_check._obj_key(d0=dref0, short=_IREF)

        # v0
        if isinstance(v0, (np.ndarray, list, tuple)):
            dref[k0] = {'data': v0}
            v0 = dref[k0]

        c0 = (
            isinstance(v0, dict)
            and (
                isinstance(v0.get('data'), (np.ndarray, list, tuple))
                or isinstance(v0.get('size'), int)
            )
        )
        if not c0:
            msg = "v0 must be a dict with either 'data' or 'size'"
            raise Exception(msg)

        dref2[key] = dict(dref[k0])

    # -----------------------
    # raise except if conflict with existing entry

    _check_conflicts(dd=dref2, dd0=dref0, dd_name='dref')

    # ----------------
    # Add size / data if relevant

    ddata_add = {
        k0: {'data': None}
        for k0, v0 in dref2.items()
        if v0.get('data') is not None
        and k0 not in ddata0.keys()
    }
    for k0, v0 in dref2.items():
        if v0.get('data') is not None:
            data, dref2[k0]['size'] = _check_dataref(
                data=v0['data'], key=k0,
            )
            if k0 in ddata_add.keys():
                ddata_add[k0]['data'] = data
                ddata_add[k0]['ref'] = (k0,)
                ddata_add[k0].update({
                    k1: v1 for k1, v1 in v0.items()
                    if k1 not in ['size', 'ldata']
                })

    # get rid of extra keys
    dref = {
        k0: {k1: v0.get(k1) for k1 in _DRESERVED_KEYS['dref']}
        for k0, v0 in dref2.items()
    }
    return dref, ddata_add


# #############################################################################
# #############################################################################
#                           ddata
# #############################################################################


def _check_data(data=None, key=None, max_ndim=None):
    """ Check the conformity of data to be a valid reference

    max_ndim allows to define a maximum number of dimensions
    lists and tuple of non-uniform len elements are converted to object arrays

    """

    # if not array
    # => try converting or get class (dict, mesh...)
    shape = None
    c0_array = (
        isinstance(data, np.ndarray)
        or scpsp.issparse(data)
    )

    # if not array => list, tuple
    if not c0_array:
        if isinstance(data, (list, tuple)):
            c0 = (
                all([hasattr(oo, '__iter__') for oo in data])
                and len(set([len(oo) for oo in data])) != 1
            )
            if c0:
                # non-uniform len of element => object array
                data = np.array(data, dtype=object)
                shape = (data.shape[0],)

            else:
                # uniform len of all elements => convert to array
                try:
                    data = np.array(data)
                    shape = data.shape
                    c0_array = True
                except Exception as err:
                    raise DataRefException(ref=key, data=data)
        else:
            msg = "Non-handled data type!"
            raise Exception(msg)

    # if array => check unique (unique + sorted)
    if shape is None:
        shape = data.shape

    # Check max_dim if any
    if c0_array and max_ndim is not None:
        if data.ndim > max_ndim:
            msg = (
                "Provided data for ddata['{key}'] has too many dimensions!\n"
                f"- ndim:     {data.ndim}\n"
                f"- max_ndim: {max_ndim}\n"
            )
            raise Exception(msg)

    # Check if valid ref candidate (monotonous = (True,))
    if c0_array:
        monotonous = tuple([
            bool(
                np.all(np.diff(data, axis=aa) > 0.)
                or np.all(np.diff(data, axis=aa) < 0.)
            )
            for aa in range(data.ndim)
        ])
    else:
        monotonous = (False,)
    return data, shape, monotonous


def _get_suitable_ref(
    shape=None,
    key=None,
    dref0=None,
    dref_add=None,
    axis=None,
):
    """  For each dimension of data.shape, identify the relevant ref index """

    if axis is None:
        axis = 0
    assert axis < len(shape)

    # list all possible ref in each dimension
    size = shape[axis]
    lref = [
        k0 for k0, v0 in dref0.items()
        if v0['size'] == size
    ]

    # perfect match
    if len(lref) == 1:
        lref = lref[0]

    # multiple matches
    elif len(lref) > 1:
        msg = (
            f"Ambiguous ref for ddata['{key}']\n"
            f"Possible matches: {lref}"
        )
        raise Exception(msg)

    # no match => create new ref
    else:
        lref0 = ds._generic_check._obj_key(d0=dref0, short=_IREF)
        lref1 = ds._generic_check._obj_key(d0=dref_add, short=_IREF)
        n0 = int(lref0[lref0[len(_IREF):]])
        n1 = int(lref1[lref1[len(_IREF):]])
        lref = lref0 if n0 > n1 else lref1

    return lref, size


def _check_data_ref(k0=None, ddata=None, dref0=None, dref_add=None):

    if dref_add is None:
        dref_add = {}

    # None => create
    if ddata[k0].get('ref') is None:
        lref = []
        for ii, ss in enumerate(ddata[k0]['shape']):
            ref, size = _get_suitable_ref(
                shape=ddata[k0]['shape'],
                key=k0,
                dref0=dref0,
                dref_add=dref_add,
                axis=ii,
            )

            if ref not in dref0.keys() and ref not in dref_add.keys():
                dref_add[ref] = {'size': size}

            lref.append(ref)

        ddata[k0]['ref'] = tuple(lref)

    # length mismatch
    elif len(ddata[k0]['shape']) != len(ddata[k0]['ref']):
        msg = (
            "Mismatching len(ref) and len(shape) for ddata['{k0}']"
        )
        raise Exception(msg)

    # length match but unknown ref or size mismatch
    else:
        for ii, ss in enumerate(ddata[k0]['shape']):
            if ddata[k0]['ref'][ii] not in dref0.keys():
                if ddata[k0]['ref'][ii] not in dref_add.keys():
                    dref_add[ddata[k0]['ref'][ii]] = {'size': ss}

            elif ss != dref0[ddata[k0]['ref'][ii]]['size']:
                msg = (
                    f"Mismatching ref size and shape for ddata['{k0}']"
                )
                raise Exception(msg)

    return dref_add


def _check_ddata(
    ddata=None,
    ddata0=None,
    dref0=None,
    reserved_keys=None,
    data_none=None,
    max_ndim=None,
):

    # ----------------
    # Trivial case

    if ddata in [None, {}]:
        return {}, {}
    if data_none is None:
        data_none = _DATA_NONE

    # ----------------
    # Check conformity

    # Basis
    # lk_opt = ['ldata', 'size', 'data']
    if not isinstance(ddata, dict):
        msg = "Arg ddata must be dict!"
        raise Exception(msg)

    ltok = (np.ndarray, list, tuple)
    lkout = [
        k0 for k0, v0 in ddata.items()
        if not (
            (k0 is None or isinstance(k0, str))
            # and k0 not in ddata0.keys()
            and (
                (isinstance(v0, ltok) or scpsp.issparse(v0))
                or (
                    isinstance(v0, dict)
                    and all([isinstance(ss, str) for ss in v0.keys()])
                    and (
                        (
                            'data' in v0.keys()
                            and (
                                isinstance(v0['data'], ltok)
                                or scpsp.issparse(v0['data'])
                            )
                            and (
                                v0.get('ref') is None
                                or isinstance(v0.get('ref'), str)
                                or (
                                    isinstance(v0.get('ref'), tuple)
                                    and all([
                                        isinstance(rr, str)
                                        for rr in v0['ref']
                                    ])
                                )
                            )
                        )
                    )
                )
            )
        )
    ]

    # Raise exception if non-conformity
    if len(lkout) != 0:
        msg = (
            """
            Arg ddata must be a dict of the form:
            dict(
                'data0': {'ref': 'ref0', 'data': list, ...},       (A)
                'data1': {'ref': ('ref0', 'ref1'), 'data': np.array, ...},  (B)
                'data2': {'data': np.array, ...},                (C)
                ...
                'datan': np.array,                               (D)
            )

            Where:
                - each 'datai' is a unique str identifier
                - (A) & (B): 'data' is provided as well as 'ref'
                - (C): 'ref' is not provided if len(self.dref) == 1
                - (D): only the data array is provided

            If ref = True, the data is itself considered a ref

            The following keys do not match the criteria:
            """
            + '\t- ' + '\n\t- '.join(lkout)
        )
        raise Exception(msg)

    # -----------------------
    # raise except if conflict with existing entry

    _check_conflicts(dd=ddata, dd0=ddata0, dd_name='ddata')

    # ----------------
    # Convert and/or add ref if necessary

    ddata2 = {}
    dref_add = {}
    for k0, v0 in ddata.items():

        key = ds._generic_check._obj_key(d0=ddata, short=_IDATA)

        # convert to dict if needed
        if not isinstance(v0, dict):
            ddata[k0] = {'data': v0}

        # check data itself
        data, shape, monotonous = _check_data(
            data=ddata[k0]['data'], key=k0, max_ndim=max_ndim,
        )
        ddata[k0]['data'] = data
        ddata[k0]['shape'] = shape
        ddata[k0]['monot'] = monotonous

        # find ref
        if isinstance(v0['ref'], str):
            ddata[k0]['ref'] = (v0['ref'],)

        _check_data_ref(
            k0=k0, ddata=ddata, dref0=dref0, dref_add=dref_add,
        )

        ddata2[key] = dict(ddata[k0])

    # ------------------
    # Check ref vs shape

    for k0, v0 in ddata2.items():
        c0 = (
            isinstance(v0['ref'], tuple)
            and all([
                (
                    ss in dref0.keys()
                    and dref0[ss]['size'] == ddata2[k0]['shape'][ii]
                )
                or (
                    ss in dref_add.keys()
                    and dref_add[ss]['size'] == ddata2[k0]['shape'][ii]
                )
                for ii, ss in enumerate(v0['ref'])
            ])
        )
        if not c0:
            msg = (
                f"ddata['{k0}']['ref'] contains unknown ref:\n"
                f"\t- ddata['{k0}']['ref'] = {v0['ref']}\n"
                f"\t- dref0.keys() = {sorted(dref0.keys())}\n"
                f"\t- dref_add.keys() = {sorted(dref_add.keys())}"
            )
            raise Exception(msg)

    return ddata2, dref_add


# #############################################################################
# #############################################################################
#                           dobj
# #############################################################################


def _check_dobj(
    dobj=None,
    dobj0=None,
):

    # ----------------
    # Trivial case
    if dobj in [None, {}]:
        return {}

    # ----------------
    # Check conformity

    # map possible non-conformities
    if not isinstance(dobj, dict):
        msg = (
            "Arg dobj must be a dict!\n"
            f"\t- Provided: {type(dobj)}"
        )
        raise Exception(msg)

    # Map possible non-conformities
    dc = {}
    dobj2 = {}
    for k0, v0 in dobj.items():

        # check types (str, dict)
        c1 = isinstance(k0, str) and isinstance(v0, dict)
        if not c1:
            dc[k0] = "type(key) != str or type(value) != dict"
            continue

        # check each key / value
        lc2 = [
            f'\t- {str(k1)}: type {type(v1)}'
            f', key already in dobj0: {k1 in dobj0.get(k0, {}).keys()}'
            for k1, v1 in v0.items()
            if not (
                (k1 is None or isinstance(k1, str))
                and isinstance(v1, dict)
                and k1 not in dobj0.get(k0, {}).keys()
            )
        ]
        if len(lc2) > 0:
            dc[k0] = (
                f"The following keys of dobj['{k0}'] are not valid:\n"
                + "\n".join(lc2)
            )
            continue

        # set None to default keys if any None
        dobj2[k0] = {}
        for k1 in v0.keys():
            key = ds._generic_check._obj_key(
                d0=dobj0.get(k0, {}), short=k0[:4],
            )
            dobj2[k0][key] = dict(dobj[k0][k1])

    # Raise Exception
    if len(dc) > 0:
        msg = (
            "The following keys of dobj are non-conform:\n"
            + "\n\n".join([f"dobj['{k0}']: {v0}" for k0, v0 in dc.items()])
        )
        raise Exception(msg)

    return dobj2


# #############################################################################
# #############################################################################
#                           Params
# #############################################################################


def _harmonize_params(
    dd=None,
    dd_name=None,
    dd_name2=None,
    dstatic=None,
    lkeys=None,
    reserved_keys=None,
    ddefparams=None,
):

    # Check inputs
    if dd_name2 is None:
        dd_name2 = dd_name
    if reserved_keys is None:
        reserved_keys = _DRESERVED_KEYS[dd_name]
    if ddefparams is None:
        ddefparams = _DDEF_PARAMS[dd_name]

    # ------------------
    # list of param keys

    # Get list of known param keys
    lparams = set(itt.chain.from_iterable([
        [k1 for k1 in v0.keys() if k1 not in reserved_keys]
        for k0, v0 in dd.items()
    ]))

    # Add arbitrary params
    if lkeys is not None:
        if isinstance(lkeys, str):
            lkeys = [lkeys]
        c0 = (
            isinstance(lkeys, list)
            and all([isinstance(ss, str) for ss in lkeys])
        )
        if not c0:
            msg = "lkeys must be a list of str!"
            raise Exception(msg)
        lparams = set(lparams).intersection(lkeys)

    # ------------------
    # dparam
    for k0, v0 in ddefparams.items():
        for k1, v1 in dd.items():
            if k0 not in v1.keys():
                dd[k1][k0] = v0[1]
            else:
                # Check type if already included
                if not isinstance(dd[k1][k0], v0[0]):
                    msg = (
                        """
                        Wrong type for parameter:
                            - type({}[{}][{}]) = {}
                        - Expected: {}
                        """.format(
                            dd_name2, k1, k0, type(dd[k1][k0]), v0[0],
                        )
                    )
                    raise Exception(msg)

    for k0 in lparams:
        for k1, v1 in dd.items():
            dd[k1][k0] = dd[k1].get(k0)

    # ------------------
    # Check against dstatic0
    lkpout = [
        (k0, (k1, v0[k1]))
        for k0, v0 in dd.items()
        if k1 in dstatic.keys()
        and any([v0[k1] not in dstatic[k1].keys() for k1 in lparams])
    ]
    if len(lkpout) > 0:
        lpu = sorted(set([pp[1][0] for pp in lkpout]))
        msg0 = '\n'.join([
            '\t- {}[{}]: {}'.format(pp[0], pp[1], pp[2]) for pp in lkpout
        ])
        msg1 = '\n'.join([
            '\t- dstatic[{}]: {}'.format(pp, dstatic[pp].keys())
            for pp in lpu
        ])
        msg = (
            """
            The following parameter have non-identified values in ref_static:
            {}

            Available values:
            {}
            """.format(msg0, msg1)
        )
        raise Exception(msg)

    return dd


def _update_dstatic0(dstatic0=None, ddata0=None, dobj0=None):
    """ Count nb. of matching ref_static in ddata and dobj """

    for k0, v0 in dstatic0.items():

        # ddata
        dd = {
            k2: np.sum([ddata0[k3].get(k0) == k2 for k3 in ddata0.keys()])
            for k2 in v0.keys()
            if any([ddata0[k3].get(k0) == k2 for k3 in ddata0.keys()])
        }
        if len(dd) > 0:
            ss = 'nb. data'
            for k2, v2 in v0.items():
                dstatic0[k0][k2][ss] = int(dd.get(k2, 0))

        # dobj
        for k1, v1 in dobj0.items():
            dd = {
                k2: np.sum([v1[k3].get(k0) == k2 for k3 in v1.keys()])
                for k2 in v0.keys()
                if any([v1[k3].get(k0) == k2 for k3 in v1.keys()])
            }
            if len(dd) > 0:
                ss = 'nb. {}'.format(k1)
                for k2, v2 in v0.items():
                    dstatic0[k0][k2][ss] = int(dd.get(k2, 0))


# #############################################################################
# #############################################################################
#                           Consistency
# #############################################################################


def _consistency(
    dobj=None, dobj0=None,
    ddata=None, ddata0=None,
    dref=None, dref0=None,
    dstatic=None, dstatic0=None,
    reserved_keys=None,
    ddefparams_data=None,
    ddefparams_obj=None,
    ddefparams_static=None,
    data_none=None,
    max_ndim=None,
):

    # --------------
    # dref
    dref, ddata_add = _check_dref(
        dref=dref, dref0=dref0, ddata0=ddata0,
    )
    if ddata_add is not None:
        if ddata is None:
            ddata = ddata_add
        else:
            ddata.update(ddata_add)
    dref0.update(dref)

    # --------------
    # dstatic
    dstatic = _check_dstatic(
        dstatic=dstatic, dstatic0=dstatic0,
    )
    for k0, v0 in dstatic.items():
        if k0 not in dstatic0.keys():
            dstatic0[k0] = v0
        else:
            dstatic0[k0].update(v0)

    # --------------
    # ddata
    ddata, dref_add = _check_ddata(
        ddata=ddata, ddata0=ddata0,
        dref0=dref0,
        reserved_keys=reserved_keys,
        data_none=data_none,
        max_ndim=max_ndim,
    )
    if dref_add is not None:
        dref0.update(dref_add)
    ddata0.update(ddata)

    # -----------------
    # dobj
    dobj = _check_dobj(
        dobj=dobj, dobj0=dobj0,
    )
    for k0, v0 in dobj.items():
        if k0 not in dobj0.keys():
            dobj0[k0] = v0
        else:
            dobj0[k0].update(v0)

    # --------------
    # params harmonization - ddata
    ddata0 = _harmonize_params(
        dd=ddata0,
        dd_name='ddata',
        dstatic=dstatic0,
        ddefparams=ddefparams_data,
        reserved_keys=reserved_keys,
    )

    # --------------
    # params harmonization - dstatic
    for k0, v0 in dstatic0.items():
        dstatic0[k0] = _harmonize_params(
            dd=v0,
            dd_name='dstatic',
            dd_name2=f'dstatic[{k0}]',
            dstatic=dstatic0,
            ddefparams=ddefparams_static.get(k0),
            reserved_keys=reserved_keys,
        )

    # --------------
    # params harmonization - dobj
    for k0, v0 in dobj0.items():
        dobj0[k0] = _harmonize_params(
            dd=v0,
            dd_name='dobj',
            dd_name2='dobj[{}]'.format(k0),
            dstatic=dstatic0,
            ddefparams=ddefparams_obj.get(k0),
            reserved_keys=reserved_keys,
        )

    # --------------
    # Complement

    # ddata0
    for k0, v0 in ddata0.items():
        if v0.get('data') is None:
            continue

    # dref0
    for k0, v0 in dref0.items():
        dref0[k0]['ldata'] = sorted(set(
            k1 for k1 in ddata0.keys()
            if ddata0[k1].get('data') is not None and k0 in ddata0[k1]['ref']
        ))
        dref0[k0]['ldata_monot'] = [
            k1 for k1 in dref0[k0]['ldata']
            if ddata0[k1]['monot'] == (True,)
        ]

    # dstatic0
    _update_dstatic0(dstatic0=dstatic0, ddata0=ddata0, dobj0=dobj0)

    # --------------
    # Check conventions

    for k0, v0 in ddata0.items():
        if v0.get('data') is None:
            continue

    return dref0, dstatic0, ddata0, dobj0


"""
    # --------------
    # params
    lparam = self._ddata['lparam']
    for kk in self._ddata['lkey']:
        for pp in self._ddata['dict'][kk].keys():
            if pp not in self._reserved_all and pp not in lparam:
                lparam.append(pp)
"""


# #############################################################################
# #############################################################################
#               Switch ref
# #############################################################################


def switch_ref(
    new_ref=None,
    ddata=None,
    dref=None,
    dobj0=None,
    dstatic0=None,
    reserved_keys=None,
    ddefparams_data=None,
    data_none=None,
    max_ndim=None,
):
    """Use the provided key as ref (if valid) """

    # Check input
    c0 = (
        new_ref in ddata.keys()
        and ddata[new_ref].get('monot') == (True,)
    )
    if not c0:
        msg = (
            "\nArg new_ref must be a key to a valid ref (monotonous)!\n"
            + "\t- Provided: {}\n\n".format(new_ref)
            + "Available valid ref candidates:\n"
            + "\t- {}".format('\n\t- '.join(list(dref.keys())))
        )
        raise Exception(msg)

    # Substitute in dref
    old_ref = ddata[new_ref]['ref'][0]
    dref[new_ref] = dict(dref[old_ref])
    del dref[old_ref]

    # substitute in ddata['ref']
    for k0, v0 in ddata.items():
        if v0.get('ref') is not None and old_ref in v0['ref']:
            new = tuple([rr for rr in v0['ref']])
            ddata[k0]['ref'] = tuple([
                new_ref if rr == old_ref else rr
                for rr in v0['ref']
            ])

    return _consistency(
        ddata=ddata, ddata0={},
        dref=dref, dref0={},
        dobj=None, dobj0=dobj0,
        dstatic=None, dstatic0=dstatic0,
        reserved_keys=None,
        ddefparams_data=ddefparams_data,
        ddefparams_obj=None,
        ddefparams_static=None,
        data_none=None,
        max_ndim=None,
    )


# #############################################################################
# #############################################################################
#               Get / set / add / remove param
# #############################################################################


def _get_param(
    dd=None, dd_name=None,
    param=None, key=None, ind=None,
    returnas=None,
):
    """ Return the array of the chosen parameter (or list of parameters)

    Can be returned as:
        - dict: {param0: {key0: values0, key1: value1...}, ...}
        - np.ndarray: {param0: np.r_[values0, value1...], ...}

    """

    # ---------------
    # Check inputs

    # Get key (which data to return param for)
    key = _ind_tofrom_key(dd=dd, key=key, ind=ind, returnas=str)

    # param
    lp = [kk for kk in list(dd.values())[0].keys() if kk != 'data']
    param = ds._generic_check._check_var_iter(
        param,
        'param',
        types=list,
        types_iter=str,
        allowed=lp,
    )

    # returnas
    returnas = ds._generic_check._check_var(
        returnas,
        'returnas',
        allowed=[np.ndarray, dict],
        default=np.ndarray,
    )

    # -------------
    # Get output

    if returnas == dict:
        out = {k0: {k1: dd[k1][k0] for k1 in key} for k0 in param}
    else:
        out = {k0: np.array([dd[k1][k0] for k1 in key]) for k0 in param}

    return out


def _set_param(
    dd=None, dd_name=None,
    param=None, value=None,
    ind=None, key=None,
    distribute=None,
):
    """ Set the value of a parameter

    values can be:
        - None
        - a unique value (int, float, bool, str, tuple) => common to all keys
        - an iterable of values (array, list) => one for each key
        - a dict of values (per key)

    A subset of keys can be chosen (ind, key, fed to self.select()) to set
    only the values of some key

    """

    # ---------------
    # Check inputs

    # Get key (which data to return param for)
    key = _ind_tofrom_key(dd=dd, key=key, ind=ind, returnas=str)

    # param
    lp = [kk for kk in list(dd.values())[0].keys() if kk != 'data']
    param = ds._generic_check._check_var(
        param,
        'param',
        types=str,
        allowed=lp,
    )

    # distribute
    if isinstance(value, (list, tuple, np.ndarray)):
        defdist = len(key) == len(value)
    else:
        defdist = False
    distribute = ds._generic_check._check_var(
        distribute,
        'distribute',
        types=bool,
        default=defdist,
    )

    # ---------------
    # Set value

    # Check value - TBC: allow list
    ltypes = [str, int, np.integer, float, np.floating, tuple]
    lc = [
        isinstance(value, tuple(ltypes)),
        isinstance(value, (list, tuple, np.ndarray))
        and distribute is False,
        isinstance(value, (list, tuple, np.ndarray))
        and np.array(value).shape[0] == len(key)
        and distribute is True,
        isinstance(value, dict)
        and all([
            kk in dd.keys()
            for kk, vv in value.items()
        ])
    ]
    if not (value is None or any(lc)):
        msg = (
            """
            Accepted types for value include:
                - None
                - {}: common to all
                - list, np.ndarray: key by key
                - dict of scalar / str

            The length of value must match the selected keys ({})
            """.format(ltypes, len(key))
        )
        raise Exception(msg)

    # Update data
    if value is None or lc[0] or lc[1]:
        for kk in key:
            dd[kk][param] = value
    elif lc[2]:
        for ii, kk in enumerate(key):
            dd[kk][param] = value[ii]
    else:
        for kk, vv in value.items():
            dd[kk][param] = vv


def _add_param(
    dd=None, dd_name=None,
    param=None, value=None,
):
    """ Add a parameter, optionnally also set its value """

    # ---------------
    # Check inputs

    # Get key (which data to return param for)
    # key = _ind_tofrom_key(dd=dd, key=key, ind=ind, returnas=str)

    # param
    lp = [kk for kk in list(dd.values())[0].keys() if kk != 'data']
    param = ds._generic_check._check_var(
        param,
        'param',
        types=str,
        excluded=lp,
    )

    # Initialize and set
    for kk in dd.keys():
        dd[kk][param] = None
    _set_param(dd=dd, param=param, value=value)


def _remove_param(dd=None, dd_name=None, param=None):
    """ Remove a parameter, none by default, all if param = 'all' """

    # Check inputs
    lp = [kk for kk in list(dd.values())[0].keys() if kk != 'data']
    if param == 'all':
        param = lp
    if isinstance(param, str):
        param = [param]
    param = ds._generic_check._check_var_iter(
        param,
        'param',
        types=list,
        types_iter=str,
        allowed=lp,
    )

    # Remove
    if param is not None:
        for pp in param:
            for k0 in dd.keys():
                del dd[k0][pp]


def _ind_tofrom_key(
    dd=None, dd_name=None,
    ind=None, key=None, returnas=int,
):
    """

    From key return ind and vice-versa

    """

    # --------------------
    # Check / format input

    lc = [ind is not None, key is not None]
    if not np.sum(lc) <= 1:
        msg = ("Args ind and key cannot be prescribed simultaneously!")
        raise Exception(msg)

    returnas = ds._generic_check._check_var(
        returnas,
        'returnas',
        allowed=[int, bool, str, 'key'],
        default='key',
    )

    # -----------------
    # Compute

    # Intialize output
    out = np.zeros((len(dd),), dtype=bool)

    # Get output
    lk = list(dd.keys())
    if lc[0]:

        # Check ind
        if not isinstance(ind, np.ndarray):
            ind = np.atleast_1d(ind).ravel()
        c0 = (
            ind.ndim == 1
            and (
                (ind.dtype == np.bool and ind.size == len(dd))
                or (
                    ind.dtype == np.int
                    and np.all(np.isfinite(ind))
                    and np.max(ind) <= len(dd)
                )
            )
        )
        if not c0:
            msg = "Arg ind must be an iterable of bool or int indices!"
            raise Exception(msg)

        # return
        out[ind] = True
        if returnas in [int, str, 'key']:
            out = out.nonzero()[0]
            if returnas in [str, 'key']:
                out = np.array(
                    [kk for ii, kk in enumerate(lk) if ii in out],
                    dtype=str,
                )

    elif lc[1]:

        # Check key
        if isinstance(key, str):
            key = [key]
        key = ds._generic_check._check_var_iter(
            key,
            'key',
            types_iter=str,
            allowed=lk,
        )

        # return
        if returnas in ['key', str]:
            out = key
        else:
            for kk in key:
                out[lk.index(kk)] = True
            if returnas == int:
                out = out.nonzero()[0]
    else:
        if returnas == bool:
            out[:] = True
        elif returnas == int:
            out = np.arange(0, len(lk))
        else:
            out = lk
    return out
